"""
pyart.correct.bias_and_noise
===================

Corrects polarimetric variables for noise

.. autosummary::
    :toctree: generated/

    correct_noise_rhohv
    correct_bias
    correct_visibility
    get_sun_hits
    get_sun_hits_psr
    get_sun_hits_ivic
    sun_retrieval
    est_rhohv_rain
    est_zdr_precip
    est_zdr_snow
    selfconsistency_bias
    selfconsistency_bias2
    selfconsistency_kdp_phidp
    get_kdp_selfcons
    _est_sun_hit_pwr_hs
    _est_sun_hit_pwr_psr
    _est_sun_hit_pwr_ivic
    _est_sun_hit_zdr
    _selfconsistency_kdp_phidp


"""

from copy import deepcopy
from warnings import warn

import numpy as np
from netCDF4 import num2date

try:
    import pysolar
    _PYSOLAR_AVAILABLE = True
except ImportError:
    _PYSOLAR_AVAILABLE = False

from ..config import get_metadata, get_field_name, get_fillvalue
from ..util import estimate_noise_hs74, estimate_noise_ivic13, ivic_pct_table
from ..util import ivic_flat_reg_var_max_table, ivic_flat_reg_wind_len_table
from ..util import ivic_snr_thr_table
from .attenuation import get_mask_fzl
from .phase_proc import smooth_masked
from ..retrieve import kdp_leastsquare_single_window
from .sunlib import sun_position_pysolar, sun_position_mfr, gas_att_sun
from .sunlib import gauss_fit, retrieval_result
from ..retrieve import get_coeff_attg
from ..filters import snr_based_gate_filter, class_based_gate_filter


def correct_noise_rhohv(radar, urhohv_field=None, snr_field=None,
                        zdr_field=None, nh_field=None, nv_field=None,
                        rhohv_field=None):
    """
    Corrects RhoHV for noise according to eq. 6 in Gourley et al. 2006.
    This correction should only be performed if noise has not been subtracted
    from the signal during the moments computation.

    Parameters
    ----------
    radar : Radar
        Radar object.
    urhohv_field : str, optional
        Name of the RhoHV uncorrected for noise field.
    snr_field, zdr_field, nh_field, nv_field : str, optional
        Names of the SNR, ZDR, horizontal channel noise in dBZ and vertical
        channel noise in dBZ used to correct RhoHV.
    rhohv_field : str, optional
        Name of the rhohv field to output.

    Returns
    -------
    rhohv : dict
        Noise corrected RhoHV field.

    References
    ----------
    Gourley et al. Data Quality of the Meteo-France C-Band Polarimetric
    Radar, JAOT, 23, 1340-1356

    """
    # parse the field parameters
    if urhohv_field is None:
        urhohv_field = get_field_name('uncorrected_cross_correlation_ratio')
    if snr_field is None:
        snr_field = get_field_name('signal_to_noise_ratio')
    if zdr_field is None:
        zdr_field = get_field_name('differential_reflectivity')
    if nh_field is None:
        nh_field = get_field_name('noisedBZ_hh')
    if nv_field is None:
        nv_field = get_field_name('noisedBZ_vv')
    if rhohv_field is None:
        rhohv_field = get_field_name('cross_correlation_ratio')

    # extract fields from radar
    radar.check_field_exists(urhohv_field)
    radar.check_field_exists(snr_field)
    radar.check_field_exists(zdr_field)
    radar.check_field_exists(nh_field)
    radar.check_field_exists(nv_field)

    urhohv = radar.fields[urhohv_field]['data']
    snrdB_h = radar.fields[snr_field]['data']
    zdrdB = radar.fields[zdr_field]['data']
    nh = radar.fields[nh_field]['data']
    nv = radar.fields[nv_field]['data']

    snr_h = np.ma.power(10., 0.1*snrdB_h)
    zdr = np.ma.power(10., 0.1*zdrdB)
    alpha = np.ma.power(10., 0.1*(nh-nv))

    rhohv_data = urhohv*np.ma.sqrt((1.+1./snr_h)*(1.+zdr/(alpha*snr_h)))
    rhohv_data[rhohv_data > 1.] = 1.

    rhohv = get_metadata(rhohv_field)
    rhohv['data'] = rhohv_data

    return rhohv


def correct_bias(radar, bias=0., field_name=None):
    """
    Corrects a radar data bias. If field name is none the correction is
    applied to horizontal reflectivity by default.

    Parameters
    ----------
    radar : Radar
        Radar object.
    bias : float, optional
        The bias magnitude.
    field_name: str, optional
        Names of the field to be corrected.

    Returns
    -------
    corrected_field : dict
        The corrected field

    """
    # parse the field parameters
    if field_name is None:
        field_name = get_field_name('reflectivity')

    # extract fields from radar
    radar.check_field_exists(field_name)
    field_data = radar.fields[field_name]['data']

    corr_field_data = field_data - bias

    if field_name.startswith('corrected_'):
        corr_field_name = field_name
    else:
        corr_field_name = 'corrected_'+field_name

    corr_field = get_metadata(corr_field_name)
    corr_field['data'] = corr_field_data

    return corr_field


def correct_visibility(radar, vis_field=None, field_name=None):
    """
    Corrects the reflectivity according to visibility.
    Applied to horizontal reflectivity by default

    Parameters
    ----------
    radar : Radar
        radar object

    vis_field : str
        the name of the visibility field

    field_name: str
        names of the field to be corrected

    Returns
    -------
    corrected_field : dict
        The corrected field

    """
    # parse the field parameters
    if vis_field is None:
        vis_field = get_field_name('visibility')
    if field_name is None:
        field_name = get_field_name('reflectivity')

    # extract fields from radar
    radar.check_field_exists(vis_field)
    vis_data = radar.fields[vis_field]['data']

    radar.check_field_exists(field_name)
    field_data = radar.fields[field_name]['data']

    corr_field_data = 10.*np.ma.log10(
        np.ma.power(10., 0.1*field_data)*100./vis_data)

    if field_name.startswith('corrected_'):
        corr_field_name = field_name
    else:
        corr_field_name = 'corrected_'+field_name

    corr_field = get_metadata(corr_field_name)
    corr_field['data'] = corr_field_data

    return corr_field


def get_sun_hits(
        radar, delev_max=2., dazim_max=2., elmin=1., rmin=50000., hmin=10000.,
        nbins_min=20, attg=None, max_std_pwr=1., max_std_zdr=1.5,
        sun_position='MF', pwrh_field=None, pwrv_field=None, zdr_field=None):
    """
    get data from suspected sun hits. The sun hits are detected using the
    Hildebrand and Sekhon noise estimation

    Parameters
    ----------
    radar : Radar
        radar object
    delev_max, dazim_max : float
        maximum difference in elevation and azimuth between sun position and
        antenna pointing
    elmin : float
        minimum radar elevation angle
    rmin : float
        minimum range from which we can look for noise [m]
    hmin : float
        minimum altitude from which we can look for noise [m]. The actual
        range min will be the minimum between rmin and the range bin higher
        than hmin.
    nbins_min : int
        Minimum number of bins with valid data to consider a ray as
        potentially sun hit
    attg : float
        gas attenuation coefficient (1-way)
    max_std_pwr : float
        Maximum standard deviation of the estimated sun power to consider the
        sun signal valid [dB]
    max_std_zdr : float
        Maximum standard deviation of the estimated sun ZDR to consider the
        sun signal valid [dB]
    sun_position : str
        The function to use to compute the sun position. Can be 'MF' or
        'pysolar'
    pwrh_field, pwrv_field, zdr_field : str
        names of the signal power in dBm for the H and V polarizations and the
        differential reflectivity

    Returns
    -------
    sun_hits : dict
        a dictionary containing information of the sun hits
    new_radar : radar object
        radar object containing sweeps that contain sun hits

    """
    # get parameters
    if attg is None:
        # assign coefficients according to radar frequency
        if (radar.instrument_parameters is not None and
                'frequency' in radar.instrument_parameters):
            attg = get_coeff_attg(
                radar.instrument_parameters['frequency']['data'][0])
        else:
            attg = 0.
            warn('Unknown 1-way gas attenuation. It will be set to 0')

    ind_rmin = np.where(radar.range['data'] > rmin)[0]
    if ind_rmin.size > 0:
        ind_rmin = ind_rmin[0]
    else:
        warn('Maximum radar range below the minimum range for sun signal' +
             ' estimation. The last '+str(2*nbins_min)+' will be inspected')
        ind_rmin = int(radar.ngates-2*nbins_min)
        if ind_rmin < 0:
            warn('Radar range too short to retrieve sun signal')
            return None, None

    # parse the field parameters
    if pwrh_field is None:
        pwrh_field = get_field_name('signal_power_hh')
    if pwrv_field is None:
        pwrv_field = get_field_name('signal_power_vv')
    if zdr_field is None:
        zdr_field = get_field_name('differential_reflectivity')

    # extract fields from radar and prepare output
    try:
        radar.check_field_exists(pwrh_field)
        pwrh = radar.fields[pwrh_field]['data']
        mask_pwrh = np.ma.getmaskarray(pwrh)
        sun_hit_h = np.ma.zeros(np.shape(pwrh), dtype=np.uint8)
        sun_hit_h[mask_pwrh] = np.ma.masked
    except KeyError:
        pwrh = None
        sun_hit_h = None

    try:
        radar.check_field_exists(pwrv_field)
        pwrv = radar.fields[pwrv_field]['data']
        mask_pwrv = np.ma.getmaskarray(pwrv)
        sun_hit_v = np.ma.zeros(np.shape(pwrv), dtype=np.uint8)
        sun_hit_v[mask_pwrv] = np.ma.masked
    except KeyError:
        pwrv = None
        sun_hit_v = None

    try:
        radar.check_field_exists(zdr_field)
        zdr = radar.fields[zdr_field]['data']
        mask_zdr = np.ma.getmaskarray(zdr)
        if pwrh is not None:
            mask_zdr = np.logical_or(mask_zdr, mask_pwrh)
        if pwrv is not None:
            mask_zdr = np.logical_or(mask_zdr, mask_pwrv)
        zdr = np.ma.masked_where(mask_zdr, zdr)
        sun_hit_zdr = np.ma.zeros(np.shape(zdr), dtype=np.uint8)
        sun_hit_zdr[mask_zdr] = np.ma.masked
    except KeyError:
        zdr = None
        sun_hit_zdr = None

    if pwrh is None and pwrv is None and zdr is None:
        return None, None

    # get time at each ray
    time = num2date(radar.time['data'], radar.time['units'],
                    radar.time['calendar'])

    sun_hits = {
        'time': [], 'ray': [], 'NPrng': [],
        'rad_el': [], 'rad_az': [], 'sun_el': [], 'sun_az': [],
        'dBm_sun_hit': [], 'std(dBm_sun_hit)': [], 'NPh': [], 'NPhval': [],
        'dBmv_sun_hit': [], 'std(dBmv_sun_hit)': [], 'NPv': [], 'NPvval': [],
        'ZDR_sun_hit': [], 'std(ZDR_sun_hit)': [], 'NPzdr': [],
        'NPzdrval': []}

    for ray in range(radar.nrays):
        if radar.elevation['data'][ray] < elmin:
            continue

        if _PYSOLAR_AVAILABLE and sun_position == 'pysolar':
            elev_sun, azim_sun = sun_position_pysolar(
                time[ray], radar.latitude['data'][0],
                radar.longitude['data'][0])
        else:
            elev_sun, azim_sun = sun_position_mfr(
                time[ray], radar.latitude['data'][0],
                radar.longitude['data'][0], refraction=True)

        if elev_sun < 0:
            continue

        # get minimum range from where to compute the sun signal and
        # minimum number of gates to consider the signal valid
        ind_hmin = np.where(radar.gate_altitude['data'][ray, :] > hmin)[0]
        if ind_hmin.size > 0:
            ind_min = np.min([ind_rmin, ind_hmin[0]])
        else:
            ind_min = ind_rmin

        nrange = len(radar.range['data'])

        delev = np.ma.abs(radar.elevation['data'][ray]-elev_sun)
        dazim = np.ma.abs(
            (radar.azimuth['data'][ray]-azim_sun) *
            np.ma.cos(elev_sun*np.pi/180.))
        if dazim > 360.:
            dazim -= 360.

        if delev > delev_max or dazim > dazim_max:
            continue

        # gas atmospheric attenuation from radar to TOA
        attg_sun = gas_att_sun(elev_sun, attg)

        sunpwrh_dBm = get_fillvalue()
        sunpwrh_std = get_fillvalue()
        sunpwrh_npoints = 0
        nvalidh = 0
        sun_hit_h_ray = None
        if pwrh is not None:
            (sunpwrh_dBm, sunpwrh_std, sunpwrh_npoints, nvalidh,
             sun_hit_h_ray) = _est_sun_hit_pwr_hs(
                 pwrh[ray, :], sun_hit_h[ray, :], attg_sun, max_std_pwr,
                 nbins_min, ind_min)
            sun_hit_h[ray, :] = sun_hit_h_ray

        sunpwrv_dBm = get_fillvalue()
        sunpwrv_std = get_fillvalue()
        sunpwrv_npoints = 0
        nvalidv = 0
        sun_hit_v_ray = None
        if pwrv is not None:
            (sunpwrv_dBm, sunpwrv_std, sunpwrv_npoints, nvalidv,
             sun_hit_v_ray) = _est_sun_hit_pwr_hs(
                 pwrv[ray, :], sun_hit_v[ray, :], attg_sun, max_std_pwr,
                 nbins_min, ind_min)
            sun_hit_v[ray, :] = sun_hit_v_ray

        sunzdr = get_fillvalue()
        sunzdr_std = get_fillvalue()
        sunzdr_npoints = 0
        nvalidzdr = 0
        if zdr is not None:
            (sunzdr, sunzdr_std, sunzdr_npoints, nvalidzdr,
             sun_hit_zdr_ray) = _est_sun_hit_zdr(
                 zdr[ray, :], sun_hit_zdr[ray, :], sun_hit_h_ray,
                 sun_hit_v_ray, max_std_zdr, nbins_min, ind_min)
            sun_hit_zdr[ray, :] = sun_hit_zdr_ray

        sun_hits['time'].append(time[ray])
        sun_hits['ray'].append(ray)
        sun_hits['NPrng'].append(nrange)
        sun_hits['rad_el'].append(radar.elevation['data'][ray])
        sun_hits['rad_az'].append(radar.azimuth['data'][ray])
        sun_hits['sun_el'].append(elev_sun)
        sun_hits['sun_az'].append(azim_sun)
        sun_hits['dBm_sun_hit'].append(sunpwrh_dBm)
        sun_hits['std(dBm_sun_hit)'].append(sunpwrh_std)
        sun_hits['NPh'].append(sunpwrh_npoints)
        sun_hits['NPhval'].append(nvalidh)
        sun_hits['dBmv_sun_hit'].append(sunpwrv_dBm)
        sun_hits['std(dBmv_sun_hit)'].append(sunpwrv_std)
        sun_hits['NPv'].append(sunpwrv_npoints)
        sun_hits['NPvval'].append(nvalidv)
        sun_hits['ZDR_sun_hit'].append(sunzdr)
        sun_hits['std(ZDR_sun_hit)'].append(sunzdr_std)
        sun_hits['NPzdr'].append(sunzdr_npoints)
        sun_hits['NPzdrval'].append(nvalidzdr)

    nhits = len(sun_hits['time'])
    if nhits == 0:
        return None, None

    # create output radar
    new_radar = deepcopy(radar)
    new_radar.fields = dict()

    if pwrh is not None:
        pwrh_dict = get_metadata(pwrh_field)
        pwrh_dict['data'] = pwrh

        sun_hit_h_dict = get_metadata('sun_hit_h')
        sun_hit_h_dict['data'] = sun_hit_h+1
        sun_hit_h_dict.update({'_FillValue': 0})

        new_radar.add_field(pwrh_field, pwrh_dict)
        new_radar.add_field('sun_hit_h', sun_hit_h_dict)

    if pwrv is not None:
        pwrv_dict = get_metadata(pwrv_field)
        pwrv_dict['data'] = pwrv

        sun_hit_v_dict = get_metadata('sun_hit_v')
        sun_hit_v_dict['data'] = sun_hit_v+1
        sun_hit_v_dict.update({'_FillValue': 0})

        new_radar.add_field(pwrv_field, pwrv_dict)
        new_radar.add_field('sun_hit_v', sun_hit_v_dict)

    if zdr is not None:
        zdr_dict = get_metadata(zdr_field)
        zdr_dict['data'] = zdr

        sun_hit_zdr_dict = get_metadata('sun_hit_zdr')
        sun_hit_zdr_dict['data'] = sun_hit_zdr+1
        sun_hit_zdr_dict.update({'_FillValue': 0})

        new_radar.add_field(zdr_field, zdr_dict)
        new_radar.add_field('sun_hit_zdr', sun_hit_zdr_dict)

    sweeps = []
    for i in range(nhits):
        for sweep in range(new_radar.nsweeps):
            ray_start, ray_end = new_radar.get_start_end(sweep)
            if ((ray_start <= sun_hits['ray'][i]) and
                    (ray_end >= sun_hits['ray'][i])):
                sweeps.append(sweep)
                break

    new_radar = new_radar.extract_sweeps(sweeps)

    # write sun hit data as ndarray
    sun_hits['ray'] = np.asarray(sun_hits['ray'])
    sun_hits['NPrng'] = np.asarray(sun_hits['NPrng'])
    sun_hits['rad_el'] = np.asarray(sun_hits['rad_el'])
    sun_hits['rad_az'] = np.asarray(sun_hits['rad_az'])
    sun_hits['sun_el'] = np.asarray(sun_hits['sun_el'])
    sun_hits['sun_az'] = np.asarray(sun_hits['sun_az'])
    sun_hits['dBm_sun_hit'] = np.ma.masked_values(
        sun_hits['dBm_sun_hit'], get_fillvalue())
    sun_hits['std(dBm_sun_hit)'] = np.ma.masked_values(
        sun_hits['std(dBm_sun_hit)'], get_fillvalue())
    sun_hits['NPh'] = np.asarray(sun_hits['NPh'])
    sun_hits['NPhval'] = np.asarray(sun_hits['NPhval'])
    sun_hits['dBmv_sun_hit'] = np.ma.masked_values(
        sun_hits['dBmv_sun_hit'], get_fillvalue())
    sun_hits['std(dBmv_sun_hit)'] = np.ma.masked_values(
        sun_hits['std(dBm_sun_hit)'], get_fillvalue())
    sun_hits['NPv'] = np.asarray(sun_hits['NPv'])
    sun_hits['NPvval'] = np.asarray(sun_hits['NPvval'])
    sun_hits['ZDR_sun_hit'] = np.ma.masked_values(
        sun_hits['ZDR_sun_hit'], get_fillvalue())
    sun_hits['std(ZDR_sun_hit)'] = np.ma.masked_values(
        sun_hits['std(ZDR_sun_hit)'], get_fillvalue())
    sun_hits['NPzdr'] = np.asarray(sun_hits['NPzdr'])
    sun_hits['NPzdrval'] = np.asarray(sun_hits['NPzdrval'])

    return sun_hits, new_radar


def get_sun_hits_psr(
        radar, delev_max=2., dazim_max=2., elmin=5.,
        noise_thres=None, attg=None, sun_position='MF',
        pwrh_field=None, pwrv_field=None):
    """
    get data from suspected sun hits. The sun hits are detected by
    a simple noise threshold from the PSR data.

    Parameters
    ----------
    radar : Radar
        radar object
    delev_max, dazim_max : float
        maximum difference in elevation and azimuth between sun position and
        antenna pointing
    elmin : float
        minimum radar elevation angle
    noise_thres : float
        noise threshold to separate valid sun samples
    attg : float
        gas attenuation coefficient (1-way)
    sun_position : str
        The function to use to compute the sun position. Can be 'MF' or
        'pysolar'
    pwrh_field, pwrv_field : str
        names of the signal power in dBm for the H and V polarizations

    Returns
    -------
    sun_hits : dict
        a dictionary containing information of the sun hits
    """
    # get parameters
    if attg is None:
        # assign coefficients according to radar frequency
        if (radar.instrument_parameters is not None and
                'frequency' in radar.instrument_parameters):
            attg = get_coeff_attg(
                radar.instrument_parameters['frequency']['data'][0])
        else:
            attg = 0.
            warn('Unknown 1-way gas attenuation. It will be set to 0')

    # parse the field parameters
    if pwrh_field is None:
        pwrh_field = get_field_name('signal_power_hh')
    if pwrv_field is None:
        pwrv_field = get_field_name('signal_power_vv')

    # extract fields from radar and prepare output
    try:
        radar.check_field_exists(pwrh_field)
        pwrh = radar.fields[pwrh_field]['data'][:, 0]
    except KeyError:
        pwrh = None

    try:
        radar.check_field_exists(pwrv_field)
        pwrv = radar.fields[pwrv_field]['data'][:, 0]
    except KeyError:
        pwrv = None

    if pwrh is None and pwrv is None:
        return None, None

    # get time at each ray
    time = num2date(radar.time['data'], radar.time['units'],
                    radar.time['calendar'])

    sun_hits = {
        'time': [], 'ray': [], 'NPrng': [], 'nhits': [],
        'rad_el': [], 'rad_az': [], 'sun_el': [], 'sun_az': [],
        'dBm_sun_hit': [], 'std(dBm_sun_hit)': [],
        'dBmv_sun_hit': [], 'std(dBmv_sun_hit)': []}

    for ray in range(radar.nrays):
        if radar.elevation['data'][ray] < elmin:
            continue

        if _PYSOLAR_AVAILABLE and sun_position == 'pysolar':
            elev_sun, azim_sun = sun_position_pysolar(
                time[ray], radar.latitude['data'][0],
                radar.longitude['data'][0])
        else:
            elev_sun, azim_sun = sun_position_mfr(
                time[ray], radar.latitude['data'][0],
                radar.longitude['data'][0], refraction=True)

        if elev_sun < 0:
            continue

        if pwrh is not None:
            if np.ma.is_masked(pwrh[ray]):
                continue
            else:
                if pwrh[ray] <= noise_thres:
                    continue

        if pwrv is not None:
            if np.ma.is_masked(pwrv[ray]):
                continue
            else:
                if pwrv[ray] <= noise_thres:
                    continue

        nrange = len(radar.range['data'])

        delev = np.ma.abs(radar.elevation['data'][ray]-elev_sun)
        dazim = np.ma.abs(
            (radar.azimuth['data'][ray]-azim_sun) *
            np.ma.cos(elev_sun*np.pi/180.))
        if dazim > 360.:
            dazim -= 360.

        if delev > delev_max or dazim > dazim_max:
            continue

        # gas atmospheric attenuation from radar to TOA
        attg_sun = gas_att_sun(elev_sun, attg)

        sunpwrh_dBm = get_fillvalue()
        if pwrh is not None:
            sunpwrh_dBm = _est_sun_hit_pwr_psr(pwrh[ray], attg_sun)

        sunpwrv_dBm = get_fillvalue()
        if pwrv is not None:
            sunpwrv_dBm = _est_sun_hit_pwr_psr(pwrv[ray], attg_sun)

        sun_hits['time'].append(time[ray])
        sun_hits['ray'].append(ray)
        sun_hits['NPrng'].append(nrange)
        sun_hits['rad_el'].append(radar.elevation['data'][ray])
        sun_hits['rad_az'].append(radar.azimuth['data'][ray])
        sun_hits['sun_el'].append(elev_sun)
        sun_hits['sun_az'].append(azim_sun)
        sun_hits['dBm_sun_hit'].append(sunpwrh_dBm)
        sun_hits['std(dBm_sun_hit)'].append(0.)
        sun_hits['dBmv_sun_hit'].append(sunpwrv_dBm)
        sun_hits['std(dBmv_sun_hit)'].append(0.)

    nhits = len(sun_hits['time'])
    if nhits == 0:
        return None, None

    # write sun hit data as ndarray
    sun_hits['ray'] = np.asarray(sun_hits['ray'])
    sun_hits['NPrng'] = np.asarray(sun_hits['NPrng'])
    sun_hits['nhits'] = np.asarray(nhits)
    sun_hits['rad_el'] = np.asarray(sun_hits['rad_el'])
    sun_hits['rad_az'] = np.asarray(sun_hits['rad_az'])
    sun_hits['sun_el'] = np.asarray(sun_hits['sun_el'])
    sun_hits['sun_az'] = np.asarray(sun_hits['sun_az'])
    sun_hits['dBm_sun_hit'] = np.ma.masked_values(
        sun_hits['dBm_sun_hit'], get_fillvalue())
    sun_hits['std(dBm_sun_hit)'] = np.ma.masked_values(
        sun_hits['std(dBm_sun_hit)'], get_fillvalue())
    sun_hits['dBmv_sun_hit'] = np.ma.masked_values(
        sun_hits['dBmv_sun_hit'], get_fillvalue())
    sun_hits['std(dBmv_sun_hit)'] = np.ma.masked_values(
        sun_hits['std(dBmv_sun_hit)'], get_fillvalue())

    return sun_hits


def get_sun_hits_ivic(
        radar, delev_max=2., dazim_max=2., elmin=1., npulses_ray=30,
        nbins_min=800, iterations=10, attg=None, sun_position='MF',
        max_std_zdr=1.5, pwrh_field=None, pwrv_field=None, zdr_field=None):

    """
    get data from suspected sun hits. The sun hits are detected using the
    Ivic et al. (2003) noise estimator

    Parameters
    ----------
    radar : Radar
        radar object
    delev_max, dazim_max : float
        maximum difference in elevation and azimuth between sun position and
        antenna pointing
    elmin : float
        minimum radar elevation angle
    npulses_ray : int
        Default number of pulses used in the computation of the ray. If the
        number of pulses is not in radar.instrument_parameters this will be
        used instead



    nbins_min: int
        minimum number of gates with noise to consider the retrieval valid
    iterations: int
        number of iterations in step 7 of Ivic algorithm
    attg : float
        gas attenuation coefficient (1-way)
    sun_position : str
        The function to use to compute the sun position. Can be 'MF' or
        'pysolar'
    max_std_zdr : float
        Maximum standard deviation of the estimated sun ZDR to consider the
        sun signal valid [dB]
    pwrh_field, pwrv_field, zdr_field : str
        names of the signal power in dBm for the H and V polarizations and the
        differential reflectivity

    Returns
    -------
    sun_hits : dict
        a dictionary containing information of the sun hits
    new_radar : radar object
        radar object containing sweeps that contain sun hits

    """
    # get parameters
    if attg is None:
        # assign coefficients according to radar frequency
        if (radar.instrument_parameters is not None and
                'frequency' in radar.instrument_parameters):
            attg = get_coeff_attg(
                radar.instrument_parameters['frequency']['data'][0])
        else:
            attg = 0.
            warn('Unknown 1-way gas attenuation. It will be set to 0')

    # parse the field parameters
    if pwrh_field is None:
        pwrh_field = get_field_name('signal_power_hh')
    if pwrv_field is None:
        pwrv_field = get_field_name('signal_power_vv')
    if zdr_field is None:
        zdr_field = get_field_name('differential_reflectivity')

    # extract fields from radar and prepare output
    try:
        radar.check_field_exists(pwrh_field)
        pwrh = radar.fields[pwrh_field]['data']
        mask_pwrh = np.ma.getmaskarray(pwrh)
        sun_hit_h = np.ma.zeros(np.shape(pwrh), dtype=np.uint8)
        sun_hit_h[mask_pwrh] = np.ma.masked
    except KeyError:
        pwrh = None
        sun_hit_h = None

    try:
        radar.check_field_exists(pwrv_field)
        pwrv = radar.fields[pwrv_field]['data']
        mask_pwrv = np.ma.getmaskarray(pwrv)
        sun_hit_v = np.ma.zeros(np.shape(pwrv), dtype=np.uint8)
        sun_hit_v[mask_pwrv] = np.ma.masked
    except KeyError:
        pwrv = None
        sun_hit_v = None

    try:
        radar.check_field_exists(zdr_field)
        zdr = radar.fields[zdr_field]['data']
        mask_zdr = np.ma.getmaskarray(zdr)
        if pwrh is not None:
            mask_zdr = np.logical_or(mask_zdr, mask_pwrh)
        if pwrv is not None:
            mask_zdr = np.logical_or(mask_zdr, mask_pwrv)
        zdr = np.ma.masked_where(mask_zdr, zdr)
        sun_hit_zdr = np.ma.zeros(np.shape(zdr), dtype=np.uint8)
        sun_hit_zdr[mask_zdr] = np.ma.masked
    except KeyError:
        zdr = None
        sun_hit_zdr = None

    if pwrh is None and pwrv is None and zdr is None:
        return None, None

    # get time at each ray
    time = num2date(radar.time['data'], radar.time['units'],
                    radar.time['calendar'])

    # get number of pulses per ray
    if radar.instrument_parameters is not None:
        if 'number_of_pulses' in radar.instrument_parameters:
            npulses = radar.instrument_parameters['number_of_pulses']['data']
        else:
            warn('Unknown number of pulses per ray. Default value ' +
                 str(npulses_ray)+' will be used for all rays')
            npulses = np.zeros(radar.nrays, dtype=int)+npulses_ray
    else:
        warn('Unknown number of pulses per ray. Default value ' +
             str(npulses_ray)+' will be used for all rays')
        npulses = np.zeros(radar.nrays, dtype=int)+npulses_ray

    if pwrh is not None or pwrv is not None:
        # threshold for step 1:
        pct = ivic_pct_table(npulses)

        # threshold for step 2:
        flat_reg_wlen = ivic_flat_reg_wind_len_table(npulses)
        # we want an odd window
        for i in range(radar.nrays):
            if flat_reg_wlen[i] % 2 == 0:
                flat_reg_wlen[i] += 1
        flat_reg_var_max = ivic_flat_reg_var_max_table(npulses)

        # threshold for step 3:
        snr_thr = ivic_snr_thr_table(npulses)

    sun_hits = {
        'time': [], 'ray': [], 'NPrng': [],
        'rad_el': [], 'rad_az': [], 'sun_el': [], 'sun_az': [],
        'dBm_sun_hit': [], 'std(dBm_sun_hit)': [], 'NPh': [], 'NPhval': [],
        'dBmv_sun_hit': [], 'std(dBmv_sun_hit)': [], 'NPv': [], 'NPvval': [],
        'ZDR_sun_hit': [], 'std(ZDR_sun_hit)': [], 'NPzdr': [],
        'NPzdrval': []}

    for ray, npuls in enumerate(npulses):
        if radar.elevation['data'][ray] < elmin:
            continue

        if _PYSOLAR_AVAILABLE and sun_position == 'pysolar':
            elev_sun, azim_sun = sun_position_pysolar(
                time[ray], radar.latitude['data'][0],
                radar.longitude['data'][0])
        else:
            elev_sun, azim_sun = sun_position_mfr(
                time[ray], radar.latitude['data'][0],
                radar.longitude['data'][0], refraction=True)

        if elev_sun < 0:
            continue

        delev = np.ma.abs(radar.elevation['data'][ray]-elev_sun)
        dazim = np.ma.abs(
            (radar.azimuth['data'][ray]-azim_sun) *
            np.ma.cos(elev_sun*np.pi/180.))
        if dazim > 360.:
            dazim -= 360.

        if delev > delev_max or dazim > dazim_max:
            continue

        # gas atmospheric attenuation from radar to TOA
        attg_sun = gas_att_sun(elev_sun, attg)

        sunpwrh_dBm = get_fillvalue()
        sunpwrh_std = get_fillvalue()
        sunpwrh_npoints = 0
        nvalidh = 0
        sun_hit_h_ray = None
        if pwrh is not None:
            (sunpwrh_dBm, sunpwrh_std, sunpwrh_npoints, nvalidh,
             sun_hit_h_ray) = _est_sun_hit_pwr_ivic(
                 pwrh[ray, :], sun_hit_h[ray, :], attg_sun, pct[ray],
                 flat_reg_wlen[ray], flat_reg_var_max[ray], snr_thr[ray], npuls,
                 nbins_min, iterations)
            sun_hit_h[ray, :] = sun_hit_h_ray

        sunpwrv_dBm = get_fillvalue()
        sunpwrv_std = get_fillvalue()
        sunpwrv_npoints = 0
        nvalidv = 0
        sun_hit_v_ray = None
        if pwrv is not None:
            (sunpwrv_dBm, sunpwrv_std, sunpwrv_npoints, nvalidv,
             sun_hit_v_ray) = _est_sun_hit_pwr_ivic(
                 pwrv[ray, :], sun_hit_v[ray, :], attg_sun, pct[ray],
                 flat_reg_wlen, flat_reg_var_max[ray], snr_thr[ray], npuls,
                 nbins_min, iterations)
            sun_hit_v[ray, :] = sun_hit_v_ray

        sunzdr = get_fillvalue()
        sunzdr_std = get_fillvalue()
        sunzdr_npoints = 0
        nvalidzdr = 0
        if zdr is not None:
            (sunzdr, sunzdr_std, sunzdr_npoints, nvalidzdr,
             sun_hit_zdr_ray) = _est_sun_hit_zdr(
                 zdr[ray, :], sun_hit_zdr[ray, :], sun_hit_h_ray,
                 sun_hit_v_ray, max_std_zdr, int(nbins_min/npuls), 0)
            sun_hit_zdr[ray, :] = sun_hit_zdr_ray

        sun_hits['time'].append(time[ray])
        sun_hits['ray'].append(ray)
        sun_hits['NPrng'].append(radar.ngates)
        sun_hits['rad_el'].append(radar.elevation['data'][ray])
        sun_hits['rad_az'].append(radar.azimuth['data'][ray])
        sun_hits['sun_el'].append(elev_sun)
        sun_hits['sun_az'].append(azim_sun)
        sun_hits['dBm_sun_hit'].append(sunpwrh_dBm)
        sun_hits['std(dBm_sun_hit)'].append(sunpwrh_std)
        sun_hits['NPh'].append(sunpwrh_npoints)
        sun_hits['NPhval'].append(nvalidh)
        sun_hits['dBmv_sun_hit'].append(sunpwrv_dBm)
        sun_hits['std(dBmv_sun_hit)'].append(sunpwrv_std)
        sun_hits['NPv'].append(sunpwrv_npoints)
        sun_hits['NPvval'].append(nvalidv)
        sun_hits['ZDR_sun_hit'].append(sunzdr)
        sun_hits['std(ZDR_sun_hit)'].append(sunzdr_std)
        sun_hits['NPzdr'].append(sunzdr_npoints)
        sun_hits['NPzdrval'].append(nvalidzdr)

    nhits = len(sun_hits['time'])
    if nhits == 0:
        return None, None

    # create output radar
    new_radar = deepcopy(radar)
    new_radar.fields = dict()

    if pwrh is not None:
        pwrh_dict = get_metadata(pwrh_field)
        pwrh_dict['data'] = pwrh

        sun_hit_h_dict = get_metadata('sun_hit_h')
        sun_hit_h_dict['data'] = sun_hit_h+1
        sun_hit_h_dict.update({'_FillValue': 0})

        new_radar.add_field(pwrh_field, pwrh_dict)
        new_radar.add_field('sun_hit_h', sun_hit_h_dict)

    if pwrv is not None:
        pwrv_dict = get_metadata(pwrv_field)
        pwrv_dict['data'] = pwrv

        sun_hit_v_dict = get_metadata('sun_hit_v')
        sun_hit_v_dict['data'] = sun_hit_v+1
        sun_hit_v_dict.update({'_FillValue': 0})

        new_radar.add_field(pwrv_field, pwrv_dict)
        new_radar.add_field('sun_hit_v', sun_hit_v_dict)

    if zdr is not None:
        zdr_dict = get_metadata(zdr_field)
        zdr_dict['data'] = zdr

        sun_hit_zdr_dict = get_metadata('sun_hit_zdr')
        sun_hit_zdr_dict['data'] = sun_hit_zdr+1
        sun_hit_zdr_dict.update({'_FillValue': 0})

        new_radar.add_field(zdr_field, zdr_dict)
        new_radar.add_field('sun_hit_zdr', sun_hit_zdr_dict)

    sweeps = []
    for i in range(nhits):
        for sweep in range(new_radar.nsweeps):
            ray_start, ray_end = new_radar.get_start_end(sweep)
            if ((ray_start <= sun_hits['ray'][i]) and
                    (ray_end >= sun_hits['ray'][i])):
                sweeps.append(sweep)
                break

    new_radar = new_radar.extract_sweeps(sweeps)

    # write sun hit data as ndarray
    sun_hits['ray'] = np.asarray(sun_hits['ray'])
    sun_hits['NPrng'] = np.asarray(sun_hits['NPrng'])
    sun_hits['rad_el'] = np.asarray(sun_hits['rad_el'])
    sun_hits['rad_az'] = np.asarray(sun_hits['rad_az'])
    sun_hits['sun_el'] = np.asarray(sun_hits['sun_el'])
    sun_hits['sun_az'] = np.asarray(sun_hits['sun_az'])
    sun_hits['dBm_sun_hit'] = np.ma.masked_values(
        sun_hits['dBm_sun_hit'], get_fillvalue())
    sun_hits['std(dBm_sun_hit)'] = np.ma.masked_values(
        sun_hits['std(dBm_sun_hit)'], get_fillvalue())
    sun_hits['NPh'] = np.asarray(sun_hits['NPh'])
    sun_hits['NPhval'] = np.asarray(sun_hits['NPhval'])
    sun_hits['dBmv_sun_hit'] = np.ma.masked_values(
        sun_hits['dBmv_sun_hit'], get_fillvalue())
    sun_hits['std(dBmv_sun_hit)'] = np.ma.masked_values(
        sun_hits['std(dBm_sun_hit)'], get_fillvalue())
    sun_hits['NPv'] = np.asarray(sun_hits['NPv'])
    sun_hits['NPvval'] = np.asarray(sun_hits['NPvval'])
    sun_hits['ZDR_sun_hit'] = np.ma.masked_values(
        sun_hits['ZDR_sun_hit'], get_fillvalue())
    sun_hits['std(ZDR_sun_hit)'] = np.ma.masked_values(
        sun_hits['std(ZDR_sun_hit)'], get_fillvalue())
    sun_hits['NPzdr'] = np.asarray(sun_hits['NPzdr'])
    sun_hits['NPzdrval'] = np.asarray(sun_hits['NPzdrval'])

    return sun_hits, new_radar


def sun_retrieval(
        az_rad, az_sun, el_rad, el_sun, sun_hit, sun_hit_std,
        az_width_co=None, el_width_co=None, az_width_cross=None,
        el_width_cross=None, is_zdr=False):
    """
    Estimates sun parameters from sun hits

    Parameters
    ----------
    az_rad, az_sun, el_rad, el_sun : float array
        azimuth and elevation values of the sun and the radar
    sun_hit : float array
        sun hit value. Either power in dBm or ZDR in dB
    sun_hit_std : float array
        standard deviation of the sun hit value in dB
    az_width_co, el_width_co, az_width_cross, el_width_cross : float
        azimuth and elevation antenna width for each channel
    is_zdr : boolean
        boolean to signal that is ZDR data

    Returns
    -------
    val, val_std : float
        retrieved value and its standard deviation
    az_bias, el_bias : float
        retrieved azimuth and elevation antenna bias respect to the sun
        position
    az_width, el_width : float
        retrieved azimuth and elevation antenna widths
    nhits : int
        number of sun hits used in the retrieval
    par : float array
        and array with the 5 parameters of the Gaussian fit

    """
    # mask non hit data
    mask = np.ma.getmaskarray(sun_hit)

    az_rad = np.ma.masked_where(mask, az_rad)
    az_sun = np.ma.masked_where(mask, az_sun)
    el_rad = np.ma.masked_where(mask, el_rad)
    el_sun = np.ma.masked_where(mask, el_sun)
    sun_hit = np.ma.masked_where(mask, sun_hit)

    az_rad = az_rad.compressed()
    az_sun = az_sun.compressed()
    el_rad = el_rad.compressed()
    el_sun = el_sun.compressed()
    sun_hit = sun_hit.compressed()

    nhits = len(sun_hit)
    if not is_zdr:
        npar = 3
        if (az_width_co is None) or (el_width_co is None):
            npar = 5

    else:
        npar = 3
        if ((az_width_co is None) or (el_width_co is None) or
                (az_width_cross is None) or (el_width_cross is None)):
            npar = 5

    if nhits < npar:
        warn('Unable to perform sun retrieval. Not enough sun hits')
        return None

    par_aux, alpha, beta = gauss_fit(
        az_rad, az_sun, el_rad, el_sun, sun_hit, npar)

    if npar == 3:
        par = np.ma.zeros(npar)
        par[0:npar-1] = par_aux
        coeff = 40.*np.ma.log10(2.)
        if is_zdr:
            par[3] = (
                coeff*(1./np.ma.power(az_width_cross, 2.) -
                       1./np.ma.power(az_width_co, 2.)))
            par[4] = (
                coeff*(1./np.ma.power(el_width_cross, 2.) -
                       1./np.ma.power(el_width_co, 2.)))
        else:
            par[3] = -np.ma.power(az_width_co, 2.)/coeff
            par[4] = -np.ma.power(el_width_co, 2.)/coeff
    else:
        par = par_aux

    if (not is_zdr) and (par[3] > 0. or par[4] > 0.):
        warn('Parameter ax and/or ay of Psun equation positive. ' +
             'Retrieval not correct')

    val, val_std, az_bias, el_bias, az_width, el_width = retrieval_result(
        sun_hit, alpha, beta, par, npar)

    return val, val_std, az_bias, el_bias, az_width, el_width, nhits, par


def est_rhohv_rain(
        radar, ind_rmin=10, ind_rmax=500, zmin=20., zmax=40., thickness=700.,
        doc=None, fzl=None, rhohv_field=None, temp_field=None,
        iso0_field=None, refl_field=None, temp_ref='temperature'):
    """
    Estimates the quantiles of RhoHV in rain for each sweep

    Parameters
    ----------
    radar : Radar
        radar object
    ind_rmin, ind_rmax : int
        Min and max range index where to look for rain
    zmin, zmax : float
        The minimum and maximum reflectivity to consider the radar bin
        suitable rain
    thickness : float
        Assumed thickness of the melting layer
    doc : float
        Number of gates at the end of each ray to to remove from the
        calculation.
    fzl : float
        Freezing layer, gates above this point are not included in the
        correction.
    temp_field, iso0_field, rhohv_field, refl_field : str
        Field names within the radar object which represent the temperature,
        the height over the iso0, co-polar correlation and reflectivity
        fields. A value of None will use the default field name as defined in
        the Py-ART configuration file.
    temp_ref : str
        the field use as reference for temperature. Can be either temperature
        or height_over_iso0

    Returns
    -------
    rhohv_rain_dict : dict
        The estimated RhoHV in rain for each sweep and metadata

    """
    if (radar.instrument_parameters is not None and
            'radar_beam_width_h' in radar.instrument_parameters):
        beamwidth = (
            radar.instrument_parameters['radar_beam_width_h']['data'][0])
    else:
        warn('Unknown radar antenna beamwidth.')
        beamwidth = None

    # parse the field parameters
    if rhohv_field is None:
        rhohv_field = get_field_name('cross_correlation_ratio')
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if temp_ref == 'temperature':
        if temp_field is None:
            temp_field = get_field_name('temperature')
    elif temp_ref == 'height_over_iso0':
        if iso0_field is None:
            iso0_field = get_field_name('height_over_iso0')

    # extract fields from radar
    radar.check_field_exists(rhohv_field)
    rhohv = deepcopy(radar.fields[rhohv_field]['data'])

    radar.check_field_exists(refl_field)
    refl = radar.fields[refl_field]['data']

    # determine the valid data (i.e. data below the melting layer)
    mask = np.ma.getmaskarray(rhohv)

    mask_fzl, _ = get_mask_fzl(
        radar, fzl=fzl, doc=doc, min_temp=0., max_h_iso0=0.,
        thickness=thickness, beamwidth=beamwidth, temp_field=temp_field,
        iso0_field=iso0_field, temp_ref=temp_ref)
    mask = np.logical_or(mask, mask_fzl)

    mask_refl = np.logical_or(np.ma.getmaskarray(refl),
                              np.logical_or(refl < zmin, refl > zmax))
    mask = np.logical_or(mask, mask_refl)

    rhohv_rain = np.ma.masked_where(mask, rhohv)
    rhohv_rain[:, 0:ind_rmin] = np.ma.masked
    rhohv_rain[:, ind_rmax:-1] = np.ma.masked

    rhohv_rain_dict = get_metadata('cross_correlation_ratio_in_rain')
    rhohv_rain_dict['data'] = rhohv_rain

    return rhohv_rain_dict


def est_zdr_precip(
        radar, ind_rmin=10, ind_rmax=500, zmin=20., zmax=22., rhohvmin=0.97,
        phidpmax=10., elmax=None, thickness=700., doc=None, fzl=None,
        zdr_field=None, rhohv_field=None, phidp_field=None, temp_field=None,
        iso0_field=None, refl_field=None, temp_ref='temperature'):
    """
    Filters out all undesired data to be able to estimate ZDR bias, either in
    moderate rain or from vertically pointing scans

    Parameters
    ----------
    radar : Radar
        radar object
    ind_rmin, ind_rmax : int
        Min and max range index where to look for rain
    zmin, zmax : float
        The minimum and maximum reflectivity to consider the radar bin
        suitable rain
    rhohvmin : float
        Minimum RhoHV to consider the radar bin suitable rain
    phidpmax : float
        Maximum PhiDP to consider the radar bin suitable rain
    elmax : float
        Maximum elevation
    thickness : float
        Assumed thickness of the melting layer
    doc : float
        Number of gates at the end of each ray to to remove from the
        calculation.
    fzl : float
        Freezing layer, gates above this point are not included in the
        correction.
    zdr_field, rhohv_field, refl_field, phidp_field, temp_field,
        iso0_field : str
        Field names within the radar object which represent the differential
        reflectivity, co-polar correlation, reflectivity, differential phase,
        temperature and height relative to the iso0 fields. A value of None
        will use the default field name as defined in the Py-ART configuration
        file.
    temp_ref : str
        the field use as reference for temperature. Can be either temperature,
        height_over_iso0, fixed_fzl or None

    Returns
    -------
    zdr_prec_dict : dict
        The ZDR data complying with specifications and metadata

    """
    if (radar.instrument_parameters is not None and
            'radar_beam_width_h' in radar.instrument_parameters):
        beamwidth = (
            radar.instrument_parameters['radar_beam_width_h']['data'][0])
    else:
        warn('Unknown radar antenna beamwidth.')
        beamwidth = None

    # parse the field parameters
    if zdr_field is None:
        zdr_field = get_field_name('differential_reflectivity')
    if rhohv_field is None:
        rhohv_field = get_field_name('cross_correlation_ratio')
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if phidp_field is None:
        phidp_field = get_field_name('differential_phase')

    if temp_ref == 'temperature':
        if temp_field is None:
            temp_field = get_field_name('temperature')
    elif temp_ref == 'height_over_iso0':
        if iso0_field is None:
            iso0_field = get_field_name('height_over_iso0')

    # extract fields from radar
    radar.check_field_exists(zdr_field)
    zdr = deepcopy(radar.fields[zdr_field]['data'])

    radar.check_field_exists(refl_field)
    refl = radar.fields[refl_field]['data']

    radar.check_field_exists(rhohv_field)
    rhohv = radar.fields[rhohv_field]['data']

    radar.check_field_exists(phidp_field)
    phidp = radar.fields[phidp_field]['data']

    # determine the valid data (i.e. data below the melting layer)
    mask = np.ma.getmaskarray(zdr)

    if temp_ref is not None:
        mask_fzl, _ = get_mask_fzl(
            radar, fzl=fzl, doc=doc, min_temp=0., max_h_iso0=0.,
            thickness=thickness, beamwidth=beamwidth, temp_field=temp_field,
            iso0_field=iso0_field, temp_ref=temp_ref)
        mask = np.logical_or(mask, mask_fzl)

    mask_refl = np.logical_or(np.ma.getmaskarray(refl),
                              np.logical_or(refl < zmin, refl > zmax))
    mask = np.logical_or(mask, mask_refl)

    mask_rhohv = np.logical_or(np.ma.getmaskarray(rhohv), rhohv < rhohvmin)
    mask = np.logical_or(mask, mask_rhohv)

    mask_phidp = np.logical_or(np.ma.getmaskarray(phidp), phidp > phidpmax)
    mask = np.logical_or(mask, mask_phidp)

    zdr_prec = np.ma.masked_where(mask, zdr)
    zdr_prec[:, 0:ind_rmin] = np.ma.masked
    zdr_prec[:, ind_rmax:-1] = np.ma.masked

    if elmax is not None:
        ind_el = np.where(radar.elevation['data'] > elmax)
        zdr_prec[ind_el, :] = np.ma.masked

    zdr_prec_dict = get_metadata('differential_reflectivity_in_precipitation')
    zdr_prec_dict['data'] = zdr_prec

    return zdr_prec_dict


def est_zdr_snow(
        radar, ind_rmin=10, ind_rmax=500, zmin=0., zmax=30., snrmin=10.,
        snrmax=50., rhohvmin=0.97, kept_values=[2], phidpmax=10., kdpmax=None,
        tempmin=None, tempmax=None, elmax=None, zdr_field=None,
        rhohv_field=None, phidp_field=None, temp_field=None, snr_field=None,
        hydro_field=None, kdp_field=None, refl_field=None):
    """
    Filters out all undesired data to be able to estimate ZDR bias in snow

    Parameters
    ----------
    radar : Radar
        radar object
    ind_rmin, ind_rmax : int
        Min and max range index where to look for snow
    zmin, zmax : float
        The minimum and maximum reflectivity to consider the radar bin
        suitable snow
    snrmin, snrmax : float
        The minimum and maximum SNR to consider the radar bin suitable snow
    rhohvmin : float
        Minimum RhoHV to consider the radar bin suitable snow
    kept_values : list of int
        The hydrometeor classification values to keep
    phidpmax : float
        Maximum PhiDP to consider the radar bin suitable snow
    kdpmax : float or None
        Maximum KDP. If not none this is the maximum KDP value to consider
        the radar bin suitable snow
    tempmin, tempmax : float or None
        If not None, the minimum and maximum temperature to consider the
        radar bin suitable snow
    elmax : float
        Maximum elevation
    zdr_field, rhohv_field, refl_field, phidp_field, kdp_field, temp_field,
    snr_field, hydro_field : str
        Field names within the radar object which represent the differential
        reflectivity, co-polar correlation, reflectivity, differential phase,
        specific differnetial phase, signal to noise ratio, hydrometeor
        classification and temperature fields. A value of None will use the
        default field name as defined in the Py-ART configuration file.

    Returns
    -------
    zdr_snow_dict : dict
        The ZDR data complying with specifications and metadata

    """
    # parse the field parameters
    if zdr_field is None:
        zdr_field = get_field_name('differential_reflectivity')
    if rhohv_field is None:
        rhohv_field = get_field_name('cross_correlation_ratio')
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if phidp_field is None:
        phidp_field = get_field_name('differential_phase')
    if temp_field is None:
        temp_field = get_field_name('temperature')
    if kdp_field is None:
        kdp_field = get_field_name('specific_differential_phase')
    if snr_field is None:
        snr_field = get_field_name('signal_to_noise_ratio')
    if hydro_field is None:
        hydro_field = get_field_name('radar_echo_classification')

    # extract fields from radar
    radar.check_field_exists(zdr_field)
    zdr = deepcopy(radar.fields[zdr_field]['data'])

    radar.check_field_exists(refl_field)
    refl = radar.fields[refl_field]['data']

    radar.check_field_exists(rhohv_field)
    rhohv = radar.fields[rhohv_field]['data']

    radar.check_field_exists(phidp_field)
    phidp = radar.fields[phidp_field]['data']

    radar.check_field_exists(hydro_field)

    # determine the valid data
    mask = np.ma.getmaskarray(zdr)

    hydro_gatefilter = class_based_gate_filter(
        radar, field=hydro_field, kept_values=kept_values)
    mask_hydro = hydro_gatefilter.gate_excluded == 1

    mask = np.logical_or(mask, mask_hydro)

    if snr_field in radar.fields:
        snr_gatefilter = snr_based_gate_filter(
            radar, snr_field=snr_field, min_snr=snrmin, max_snr=snrmax)
        mask_snr = snr_gatefilter.gate_excluded == 1

        mask = np.logical_or(mask, mask_snr)
    else:
        warn('No filtering according to SNR. SNR field not available')

    mask_refl = np.logical_or(np.ma.getmaskarray(refl),
                              np.logical_or(refl < zmin, refl > zmax))
    mask = np.logical_or(mask, mask_refl)

    mask_rhohv = np.logical_or(np.ma.getmaskarray(rhohv), rhohv < rhohvmin)
    mask = np.logical_or(mask, mask_rhohv)

    mask_phidp = np.logical_or(np.ma.getmaskarray(phidp), phidp > phidpmax)
    mask = np.logical_or(mask, mask_phidp)

    if (tempmin is not None and tempmax is not None and
            temp_field in radar.fields):
        temp = radar.fields[temp_field]['data']
        mask_temp = np.logical_or(temp < tempmin, temp > tempmax)
        mask = np.logical_or(mask, mask_temp)

    if kdpmax is not None and kdp_field in radar.fields:
        kdp = radar.fields[kdp_field]['data']
        mask = np.logical_or(mask, kdp > kdpmax)

    zdr_snow = np.ma.masked_where(mask, zdr)
    zdr_snow[:, 0:ind_rmin] = np.ma.masked
    zdr_snow[:, ind_rmax:-1] = np.ma.masked

    if elmax is not None:
        ind_el = np.where(radar.elevation['data'] > elmax)
        zdr_snow[ind_el, :] = np.ma.masked

    zdr_snow_dict = get_metadata('differential_reflectivity_in_snow')
    zdr_snow_dict['data'] = zdr_snow

    return zdr_snow_dict


def selfconsistency_bias(
        radar, zdr_kdpzh_dict, min_rhohv=0.92, filter_rain=True, max_phidp=20.,
        smooth_wind_len=5, doc=None, fzl=None, thickness=700., min_rcons=20,
        dphidp_min=2, dphidp_max=16, parametrization='None', refl_field=None,
        phidp_field=None, zdr_field=None, temp_field=None, iso0_field=None,
        hydro_field=None, rhohv_field=None, temp_ref='temperature',
        check_wet_radome=True, wet_radome_refl=25., wet_radome_len_min=4,
        wet_radome_len_max=8, wet_radome_ngates_min=180,
        valid_gates_only=False, keep_points=False, kdp_wind_len=12):
    """
    Estimates reflectivity bias at each ray using the self-consistency
    algorithm by Gourley

    Parameters
    ----------
    radar : Radar
        radar object
    zdr_kdpzh_dict : dict
        dictionary containing a look up table relating ZDR with KDP/Zh for
        different elevations
    min_rhohv : float
        minimum RhoHV value to consider the data valid
    filter_rain : bool
        If True the hydrometeor classification is going to be used to filter
        out all gates that are not rain
    max_phidp : float
        maximum PhiDP value to consider the data valid
    smooth_wind_len : int
        length of the smoothing window
    doc : float
        Number of gates at the end of each ray to to remove from the
        calculation.
    fzl : float
        Freezing layer, gates above this point are not included in the
        correction.
    thickness : float
        assumed melting layer thickness [m]
    min_rcons : int
        minimum number of consecutive gates to consider a valid segment of
        PhiDP
    dphidp_min : float
        minimum differential phase shift in a segment
    dphidp_max : float
        maximum differential phase shift in a segment
    parametrization : str
        The type of parametrization for the self-consistency curves. Can be
        'None', 'Gourley', 'Wolfensberger', 'Louf', 'Gorgucci' or 'Vaccarono'.
        'None' will use tables contained in zdr_kdpzh_dict.
    refl_field, phidp_field, zdr_field : str
        Field names within the radar object which represent the reflectivity,
        differential phase and differential reflectivity fields. A value of
        None will use the default field name as defined in the Py-ART
        configuration file.
    temp_field, iso0_field, hydro_field, rhohv_field : str
        Field names within the radar object which represent the temperature,
        the height relative to the iso0, the hydrometeor classification and
        the co-polar correlation fields. A value of None will use the default
        field name as defined in the Py-ART configuration file. They are going
        to be used only if available.
    kdpsim_field, phidpsim_field : str
        Field names which represent the estimated specific differential phase
        and differential phase. A value of None will use the default
        field name as defined in the Py-ART configuration file.
    temp_ref : str
        the field use as reference for temperature. Can be either temperature,
        height_over_iso0 or fixed_fzl
    check_wet_radome : Bool
        if True the average reflectivity of the closest gates to the radar
        is going to be check to find out whether there is rain over the
        radome. If there is rain no bias will be computed
    wet_radome_refl : Float
        Average reflectivity of the gates close to the radar to consider
        the radome as wet
    wet_radome_len_min, wet_radome_len_max : int
        Mim and max gate indices of the disk around the radome used to decide
        whether the radome is wet
    wet_radome_ngates_min : int
        Minimum number of valid gates to consider that the radome is wet
    valid_gates_only : Bool
        If True the reflectivity bias obtained for each valid ray is going to
        be assigned only to gates of the segment used. That will give more
        weight to longer segments when computing the total bias.
    keep_points : Bool
        If True the ZDR, ZH and KDP of the gates used in the self-
        consistency algorithm are going to be stored for further analysis
    kdp_wind_len : int
        The length of the window used to compute KDP with the single
        window least square method

    Returns
    -------
    refl_bias_dict : dict
        the bias at each ray field and metadata
    selfconsistency_dict : dict
        If keep_poinst set, a dictionary containing the measured valid values
        of ZDR, Zh and KDP. None otherwise

    """
    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if zdr_field is None:
        zdr_field = get_field_name('zdr')
    if phidp_field is None:
        # use corrrected_differential_phase or unfolded_differential_phase
        # fields if they are available, if not use differential_phase field
        phidp_field = get_field_name('corrected_differential_phase')
        if phidp_field not in radar.fields:
            phidp_field = get_field_name('unfolded_differential_phase')
        if phidp_field not in radar.fields:
            phidp_field = get_field_name('differential_phase')
    if rhohv_field is None:
        rhohv_field = get_field_name('cross_correlation_ratio')
    if hydro_field is None:
        hydro_field = get_field_name('radar_echo_classification')

    if temp_ref == 'temperature':
        if temp_field is None:
            temp_field = get_field_name('temperature')
    elif temp_ref == 'height_over_iso0':
        if iso0_field is None:
            iso0_field = get_field_name('height_over_iso0')

    # extract fields from radar, refl, zdr and phidp must exist
    radar.check_field_exists(refl_field)
    refl = radar.fields[refl_field]['data']

    radar.check_field_exists(zdr_field)
    zdr = radar.fields[zdr_field]['data']

    radar.check_field_exists(phidp_field)
    phidp = radar.fields[phidp_field]['data']

    rhohv = None
    if min_rhohv is not None:
        try:
            radar.check_field_exists(rhohv_field)
            rhohv = radar.fields[rhohv_field]['data']
        except KeyError:
            min_rhohv = None
            rhohv = None

    hydro = None
    if filter_rain:
        try:
            radar.check_field_exists(hydro_field)
            hydro = radar.fields[hydro_field]['data']
        except KeyError:
            filter_rain = False
            hydro = None

    _, phidp_sim = _selfconsistency_kdp_phidp(
        radar, refl, zdr, phidp, zdr_kdpzh_dict, max_phidp=max_phidp,
        smooth_wind_len=smooth_wind_len, rhohv=rhohv, min_rhohv=min_rhohv,
        hydro=hydro, filter_rain=filter_rain, doc=doc, fzl=fzl,
        thickness=thickness, parametrization=parametrization,
        temp_field=temp_field, iso0_field=iso0_field, temp_ref=temp_ref)

    refl_bias = np.ma.masked_all((radar.nrays, radar.ngates))
    # check if there is rain over the radome
    if check_wet_radome:
        # self_mask = np.ma.getmaskarray(phidp_sim)[:, 0:wet_radome_len]
        refl_radome = refl[:, wet_radome_len_min:wet_radome_len_max+1]
        # refl_radome[self_mask] = np.ma.masked
        refl_avg = np.ma.mean(refl_radome)
        ngates_wet = (refl_radome.compressed()).size
        if refl_avg > wet_radome_refl and ngates_wet > wet_radome_ngates_min:
            warn('Rain over radome!!!\n Avg reflectivity between ' +
                 str(radar.range['data'][wet_radome_len_min])+' and ' +
                 str(radar.range['data'][wet_radome_len_max])+' km ' +
                 str(refl_avg)+'. Number of wet gates '+str(ngates_wet))
            refl_bias_dict = get_metadata('reflectivity_bias')
            refl_bias_dict['data'] = refl_bias
            return refl_bias_dict, None

        warn('Avg reflectivity between ' +
             str(radar.range['data'][wet_radome_len_min])+' and ' +
             str(radar.range['data'][wet_radome_len_max])+' km ' +
             str(refl_avg)+'. Number of wet gates '+str(ngates_wet))

    if keep_points:
        kdp = kdp_leastsquare_single_window(
            radar, wind_len=kdp_wind_len, min_valid=int(kdp_wind_len/2.),
            phidp_field=phidp_field, kdp_field=None, vectorize=True)
        sm_refl = smooth_masked(refl, wind_len=smooth_wind_len, min_valid=1,
                                wind_type='mean')
        sm_refl_lin = np.ma.power(10., 0.1*sm_refl)

        sm_zdr = smooth_masked(zdr, wind_len=smooth_wind_len, min_valid=1,
                               wind_type='mean')
        zdr_list = []
        zh_list = []
        kdp_list = []

    for ray in range(radar.nrays):
        # split ray in consecutive valid range bins
        isprec = np.ma.getmaskarray(phidp_sim[ray, :]) == 0
        ind_prec = np.where(isprec)[0]
        cons_list = np.split(ind_prec, np.where(np.diff(ind_prec) != 1)[0]+1)

        # check if there is a cell long enough
        found_cell = False
        for ind_prec_cell in cons_list:
            if len(ind_prec_cell) >= min_rcons:
                found_cell = True
                break
        if not found_cell:
            continue

        # check if increase in phidp is within limits and compute reflectivity
        # bias
        dphidp_obs = (
            phidp[ray, ind_prec_cell[-1]]-phidp[ray, ind_prec_cell[0]])
        if dphidp_obs < dphidp_min:
            continue

        for i in range(len(ind_prec_cell)):
            dphidp_obs = (
                phidp[ray, ind_prec_cell[-1-i]]-phidp[ray, ind_prec_cell[0]])
            if dphidp_obs > dphidp_max:
                continue

            dphidp_sim = (phidp_sim[ray, ind_prec_cell[-1-i]] -
                          phidp_sim[ray, ind_prec_cell[0]])

            if valid_gates_only:
                refl_bias[ray, ind_prec_cell[0]:ind_prec_cell[-1-i]+1] = (
                    10.*np.ma.log10(dphidp_sim/dphidp_obs))
            else:
                refl_bias[ray, 0] = 10.*np.ma.log10(dphidp_sim/dphidp_obs)

            if keep_points:
                zdr_list.extend(
                    sm_zdr[ray, ind_prec_cell[0]:ind_prec_cell[-1-i]+1])
                kdp_list.extend(
                    kdp['data'][ray, ind_prec_cell[0]:ind_prec_cell[-1-i]+1])
                zh_list.extend(
                    sm_refl_lin[ray, ind_prec_cell[0]:ind_prec_cell[-1-i]+1])
            break

    refl_bias_dict = get_metadata('reflectivity_bias')
    refl_bias_dict['data'] = refl_bias

    selfconsistency_dict = None
    if keep_points:
        selfconsistency_dict = {
            'zdr': zdr_list,
            'kdp': kdp_list,
            'zh': zh_list,
            }

    return refl_bias_dict, selfconsistency_dict


def selfconsistency_bias2(
        radar, zdr_kdpzh_dict, min_rhohv=0.92, min_zdr=0.2, filter_rain=True,
        max_phidp=20., smooth_wind_len=5, doc=None, fzl=None, thickness=700.,
        min_rcons=20, parametrization='None', refl_field=None,
        phidp_field=None, zdr_field=None, temp_field=None, iso0_field=None,
        hydro_field=None, rhohv_field=None, kdp_field=None,
        temp_ref='temperature', check_wet_radome=True, wet_radome_refl=25.,
        wet_radome_len_min=4, wet_radome_len_max=8, wet_radome_ngates_min=180,
        keep_points=False, bias_per_gate=False):
    """
    Estimates reflectivity bias at each ray using the self-consistency
    algorithm by Gourley

    Parameters
    ----------
    radar : Radar
        radar object
    zdr_kdpzh_dict : dict
        dictionary containing a look up table relating ZDR with KDP/Zh for
        different elevations
    min_rhohv : float
        minimum RhoHV value to consider the data valid
    min_zdr : float
        minimum ZDR value to consider the data valid
    filter_rain : bool
        If True the hydrometeor classification is going to be used to filter
        out all gates that are not rain
    max_phidp : float
        maximum PhiDP value to consider the data valid
    smooth_wind_len : int
        length of the smoothing window
    doc : float
        Number of gates at the end of each ray to to remove from the
        calculation.
    fzl : float
        Freezing layer, gates above this point are not included in the
        correction.
    thickness : float
        assumed melting layer thickness [m]
    min_rcons : int
        minimum number of consecutive gates to consider a valid segment of
        PhiDP
    parametrization : str
        The type of parametrization for the self-consistency curves. Can be
        'None', 'Gourley', 'Wolfensberger', 'Louf', 'Gorgucci' or 'Vaccarono'.
        'None' will use tables contained in zdr_kdpzh_dict.
    refl_field, kdp_field, zdr_field : str
        Field names within the radar object which represent the reflectivity,
        differential phase and differential reflectivity fields. A value of
        None will use the default field name as defined in the Py-ART
        configuration file.
    temp_field, iso0_field, hydro_field, rhohv_field, phidp_field : str
        Field names within the radar object which represent the temperature,
        the height relative to the iso0, the hydrometeor classification and
        the co-polar correlation fields. A value of None will use the default
        field name as defined in the Py-ART configuration file. They are going
        to be used only if available.
    kdpsim_field, phidpsim_field : str
        Field names which represent the estimated specific differential phase
        and differential phase. A value of None will use the default
        field name as defined in the Py-ART configuration file.
    temp_ref : str
        the field use as reference for temperature. Can be either temperature,
        height_over_iso0 or fixed_fzl
    check_wet_radome : Bool
        if True the average reflectivity of the closest gates to the radar
        is going to be check to find out whether there is rain over the
        radome. If there is rain no bias will be computed
    wet_radome_refl : Float
        Average reflectivity of the gates close to the radar to consider
        the radome as wet
    wet_radome_len_min, wet_radome_len_max : int
        Mim and max gate indices of the disk around the radome used to decide
        whether the radome is wet
    wet_radome_ngates_min : int
        Minimum number of valid gates to consider that the radome is wet
    keep_points : Bool
        If True the ZDR, ZH and KDP of the gates used in the self-
        consistency algorithm are going to be stored for further analysis
    bias_per_gate : Bool
        If True the bias per gate will be computed

    Returns
    -------
    kdp_data_dict : dict
        A dictionary containing valid observed and estimated using self-
        consistency values of KDP
    refl_bias_dict : dict
        If bias_per_gate is set, the bias at each gate field and metadata.
        None otherwise
    selfconsistency_dict : dict
        If keep_poinst set, a dictionary containing the measured valid values
        of ZDR, Zh and KDP. None otherwise

    """
    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if zdr_field is None:
        zdr_field = get_field_name('zdr')
    if phidp_field is None:
        # use corrrected_differential_phase or unfolded_differential_phase
        # fields if they are available, if not use differential_phase field
        phidp_field = get_field_name('corrected_differential_phase')
        if phidp_field not in radar.fields:
            phidp_field = get_field_name('unfolded_differential_phase')
        if phidp_field not in radar.fields:
            phidp_field = get_field_name('differential_phase')
    if kdp_field is None:
        kdp_field = get_field_name('specific_differential_phase')
    if rhohv_field is None:
        rhohv_field = get_field_name('cross_correlation_ratio')
    if hydro_field is None:
        hydro_field = get_field_name('radar_echo_classification')

    if temp_ref == 'temperature':
        if temp_field is None:
            temp_field = get_field_name('temperature')
    elif temp_ref == 'height_over_iso0':
        if iso0_field is None:
            iso0_field = get_field_name('height_over_iso0')

    # extract fields from radar, refl, zdr and kdp must exist
    radar.check_field_exists(refl_field)
    refl = radar.fields[refl_field]['data']

    radar.check_field_exists(zdr_field)
    zdr = radar.fields[zdr_field]['data']

    radar.check_field_exists(kdp_field)
    kdp = radar.fields[kdp_field]['data']

    phidp = None
    if max_phidp is not None:
        try:
            radar.check_field_exists(phidp_field)
            phidp = radar.fields[phidp_field]['data']
        except KeyError:
            max_phidp = None
            phidp = None

    rhohv = None
    if min_rhohv is not None:
        try:
            radar.check_field_exists(rhohv_field)
            rhohv = radar.fields[rhohv_field]['data']
        except KeyError:
            min_rhohv = None
            rhohv = None

    hydro = None
    if filter_rain:
        try:
            radar.check_field_exists(hydro_field)
            hydro = radar.fields[hydro_field]['data']
        except KeyError:
            filter_rain = False
            hydro = None

    kdp_sim_aux, _ = _selfconsistency_kdp_phidp(
        radar, refl, zdr, phidp, zdr_kdpzh_dict, max_phidp=max_phidp,
        smooth_wind_len=smooth_wind_len, rhohv=rhohv, min_rhohv=min_rhohv,
        min_zdr=min_zdr, hydro=hydro, filter_rain=filter_rain, doc=doc,
        fzl=fzl, thickness=thickness, parametrization=parametrization,
        temp_field=temp_field, iso0_field=iso0_field, temp_ref=temp_ref)

    if bias_per_gate:
        refl_bias = np.ma.masked_all((radar.nrays, radar.ngates))
    # check if there is rain over the radome
    if check_wet_radome:
        # self_mask = np.ma.getmaskarray(phidp_sim)[:, 0:wet_radome_len]
        refl_radome = refl[:, wet_radome_len_min:wet_radome_len_max+1]
        # refl_radome[self_mask] = np.ma.masked
        refl_avg = np.ma.mean(refl_radome)
        ngates_wet = (refl_radome.compressed()).size
        if refl_avg > wet_radome_refl and ngates_wet > wet_radome_ngates_min:
            warn('Rain over radome!!!\n Avg reflectivity between ' +
                 str(radar.range['data'][wet_radome_len_min])+' and ' +
                 str(radar.range['data'][wet_radome_len_max])+' km ' +
                 str(refl_avg)+'. Number of wet gates '+str(ngates_wet))

            if bias_per_gate:
                refl_bias_dict = get_metadata('reflectivity_bias')
                refl_bias_dict['data'] = refl_bias
                return None, refl_bias_dict, None

            return None, None, None

        warn('Avg reflectivity between ' +
             str(radar.range['data'][wet_radome_len_min])+' and ' +
             str(radar.range['data'][wet_radome_len_max])+' km ' +
             str(refl_avg)+'. Number of wet gates '+str(ngates_wet))

    if keep_points:
        sm_refl = smooth_masked(refl, wind_len=smooth_wind_len, min_valid=1,
                                wind_type='mean')
        sm_refl_lin = np.ma.power(10., 0.1*sm_refl)

        sm_zdr = smooth_masked(zdr, wind_len=smooth_wind_len, min_valid=1,
                               wind_type='mean')
        zdr_list = []
        zh_list = []
        kdp_list = []

    kdp_sim = []
    kdp_obs = []
    for ray in range(radar.nrays):
        # split ray in consecutive valid range bins
        isprec = np.ma.getmaskarray(kdp_sim_aux[ray, :]) == 0
        ind_prec = np.where(isprec)[0]
        cons_list = np.split(ind_prec, np.where(np.diff(ind_prec) != 1)[0]+1)

        # check if there is a cell long enough
        for ind_prec_cell in cons_list:
            if len(ind_prec_cell) < min_rcons:
                continue

            kdp_sim.extend(
                kdp_sim_aux[ray, ind_prec_cell[0]:ind_prec_cell[-1]+1])
            kdp_obs.extend(kdp[ray, ind_prec_cell[0]:ind_prec_cell[-1]+1])

            if keep_points:
                zdr_list.extend(
                    sm_zdr[ray, ind_prec_cell[0]:ind_prec_cell[-1]+1])
                kdp_list.extend(
                    kdp[ray, ind_prec_cell[0]:ind_prec_cell[-1]+1])
                zh_list.extend(
                    sm_refl_lin[ray, ind_prec_cell[0]:ind_prec_cell[-1]+1])

            if bias_per_gate:
                refl_bias[ray, ind_prec_cell[0]:ind_prec_cell[-1]+1] = (
                    10.*np.ma.log10(
                        kdp_sim_aux[ray, ind_prec_cell[0]:ind_prec_cell[-1]+1]
                        / kdp[ray, ind_prec_cell[0]:ind_prec_cell[-1]+1]))

    kdp_data_dict = {
        'kdp_sim': kdp_sim,
        'kdp_obs': kdp_obs
    }
    refl_bias_dict = None
    if bias_per_gate:
        refl_bias_dict = get_metadata('reflectivity_bias')
        refl_bias_dict['data'] = refl_bias

    selfconsistency_dict = None
    if keep_points:
        selfconsistency_dict = {
            'zdr': zdr_list,
            'kdp': kdp_list,
            'zh': zh_list,
            }

    return kdp_data_dict, refl_bias_dict, selfconsistency_dict


def selfconsistency_kdp_phidp(
        radar, zdr_kdpzh_dict, min_rhohv=0.92, filter_rain=True,
        max_phidp=20., smooth_wind_len=5, doc=None, fzl=None, thickness=700.,
        parametrization='None', refl_field=None, phidp_field=None,
        zdr_field=None, temp_field=None, iso0_field=None, hydro_field=None,
        rhohv_field=None, kdpsim_field=None, phidpsim_field=None,
        temp_ref='temperature'):
    """
    Estimates KDP and PhiDP in rain from  Zh and ZDR using a selfconsistency
    relation between ZDR, Zh and KDP. Private method

    Parameters
    ----------
    radar : Radar
        radar object
    zdr_kdpzh_dict : dict
        dictionary containing a look up table relating ZDR with KDP/Zh for
        different elevations
    min_rhohv : float
        minimum RhoHV value to consider the data valid
    filter_rain : bool
        If True the hydrometeor classification is going to be used to filter
        out all gates that are not rain
    max_phidp : float
        maximum PhiDP value to consider the data valid
    smooth_wind_len : int
        length of the smoothing window
    doc : float
        Number of gates at the end of each ray to to remove from the
        calculation.
    fzl : float
        Freezing layer, gates above this point are not included in the
        correction.
    thickness : float
        assumed melting layer thickness [m]
    parametrization : str
        The type of parametrization for the self-consistency curves. Can be
        'None', 'Gourley', 'Wolfensberger', 'Louf', 'Gorgucci' or 'Vaccarono'.
        'None' will use tables contained in zdr_kdpzh_dict.
    refl_field, phidp_field, zdr_field : str
        Field names within the radar object which represent the reflectivity,
        differential phase and differential reflectivity fields. A value of
        None will use the default field name as defined in the Py-ART
        configuration file.
    temp_field, iso0_field, hydro_field, rhohv_field : str
        Field names within the radar object which represent the temperature,
        the height relative to the iso0, the hydrometeor classification and
        the co-polar correlation fields. A value of None will use the default
        field name as defined in the Py-ART configuration file. They are going
        to be used only if available.
    kdpsim_field, phidpsim_field : str
        Field names which represent the estimated specific differential phase
        and differential phase. A value of None will use the default
        field name as defined in the Py-ART configuration file.
    temp_ref : str
        the field use as reference for temperature. Can be either temperature,
        height_over_iso0 or fixed_fzl

    Returns
    -------
    kdp_sim_dict, phidp_sim_dict : dict
        the KDP and PhiDP estimated fields and metadata

    """
    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if zdr_field is None:
        zdr_field = get_field_name('zdr')
    if phidp_field is None:
        # use corrrected_differential_phase or unfolded_differential_phase
        # fields if they are available, if not use differential_phase field
        phidp_field = get_field_name('corrected_differential_phase')
        if phidp_field not in radar.fields:
            phidp_field = get_field_name('unfolded_differential_phase')
        if phidp_field not in radar.fields:
            phidp_field = get_field_name('differential_phase')
    if rhohv_field is None:
        rhohv_field = get_field_name('cross_correlation_ratio')
    if hydro_field is None:
        hydro_field = get_field_name('radar_echo_classification')
    if kdpsim_field is None:
        kdpsim_field = get_field_name('specific_differential_phase')
    if phidpsim_field is None:
        phidpsim_field = get_field_name('differential_phase')

    if temp_ref == 'temperature':
        if temp_field is None:
            temp_field = get_field_name('temperature')
    elif temp_ref == 'height_over_iso0':
        if iso0_field is None:
            iso0_field = get_field_name('height_over_iso0')
    elif temp_ref == 'hydroclass':
        if hydro_field is None:
            hydro_field = get_field_name('radar_echo_classification')

    # extract fields from radar, refl, zdr and phidp must exist
    radar.check_field_exists(refl_field)
    refl = radar.fields[refl_field]['data']

    radar.check_field_exists(zdr_field)
    zdr = radar.fields[zdr_field]['data']

    radar.check_field_exists(phidp_field)
    phidp = radar.fields[phidp_field]['data']

    if min_rhohv is not None:
        try:
            radar.check_field_exists(rhohv_field)
            rhohv = radar.fields[rhohv_field]['data']
        except KeyError:
            min_rhohv = None
            rhohv = None

    if filter_rain:
        try:
            radar.check_field_exists(hydro_field)
            hydro = radar.fields[hydro_field]['data']
        except KeyError:
            filter_rain = False
            hydro = None

    kdp_sim, phidp_sim = _selfconsistency_kdp_phidp(
        radar, refl, zdr, phidp, zdr_kdpzh_dict, max_phidp=max_phidp,
        smooth_wind_len=smooth_wind_len, rhohv=rhohv, min_rhohv=min_rhohv,
        hydro=hydro, filter_rain=filter_rain, doc=doc, fzl=fzl,
        thickness=thickness, parametrization=parametrization,
        temp_field=temp_field, iso0_field=iso0_field, temp_ref=temp_ref)

    kdp_sim_dict = get_metadata(kdpsim_field)
    kdp_sim_dict['data'] = kdp_sim

    phidp_sim_dict = get_metadata(phidpsim_field)
    phidp_sim_dict['data'] = phidp_sim

    return kdp_sim_dict, phidp_sim_dict


def get_kdp_selfcons(zdr, refl, ele_vec, zdr_kdpzh_dict,
                     parametrization='None'):
    """
    Estimates KDP and PhiDP in rain from  Zh and ZDR using a selfconsistency
    relation between ZDR, Zh and KDP

    Parameters
    ----------
    zdr, refl : ndarray 2D
        reflectivity and differential reflectivity fields
    ele_vec : ndarray 1D
        vector containing the elevation angles of each ray
    zdr_kdpzh_dict : dict
        dictionary containing a look up table relating ZDR with KDP/Zh for
        different elevations
    parametrization : str
        The type of parametrization for the self-consistency curves. Can be
        'None', 'Gourley', 'Wolfensberger', 'Louf', 'Gorgucci' or 'Vaccarono'.
        'None' will use tables contained in zdr_kdpzh_dict. The parametrized
        curves are obtained from literature except for Wolfensberger that was
        derived from disdrometer data obtained by MeteoSwiss and EPFL. All
        parametrizations are valid for C-band only except that of Gourley.

    Returns
    -------
    kdp_sim : ndarray 2D
        the KDP estimated from zdr and refl

    References
    ----------
    E. Gorgucci, G. Scarchilli, V. Chandrasekar, "Calibration of radars using
    polarimetric techniques", IEEE Transactions on Geoscience and Remote
    Sensing, 1992, 30

    J.J. Gourley, A.J. Illingworth, P. Tabary, "Absolute Calibration of Radar
    Reflectivity Using Redundancy of the Polarization Observations and Implied
    Constraints on Drop Shapes", J. of Atmospheric and Oceanic Technology,
    2009, 26

    V. Louf, A. Protat, R.A. Warren, S.M. Collis, D.B. Wolff, S. Raunyiar,
    C. Jakob, W. A. Petersen, "An Integrated Approach to Weather Radar
    Calibration and Monitoring Using Ground Clutter and Satellite
    Comparisons", J. of Atmospheric and Oceanic Technology, 2019, 36

    M. Vaccarono, R. Bechini, C. V. Chandrasekar, R. Cremonini, C. Cassardo,
    "An integrated approach to monitoring the calibration stability of
    operational dual-polarization radars", Atmos. Meas. Tech., 2016, 9

    """
    # prepare output
    kdpzh = np.ma.masked_all(np.shape(zdr))

    refl_lin = np.ma.power(10., refl/10.)
    zdr_mask = np.ma.getmaskarray(zdr)

    if parametrization == 'None':
        # process each elevation in the look up table present in the field
        ele_rounded = ele_vec.astype(int)
        for i in range(len(zdr_kdpzh_dict['elev'])):
            # look for gates with valid elevation
            ind_ray = np.where(ele_rounded == zdr_kdpzh_dict['elev'][i])[0]
            if ind_ray.size == 0:
                continue

            # look for valid ZDR
            zdr_valid = zdr[ind_ray, :].compressed()
            if zdr_valid.size == 0:
                continue

            # auxiliary array with the size of valid rays
            kdpzh_aux = np.ma.masked_all(np.size(zdr[ind_ray, :]))

            mask_aux = zdr_mask[ind_ray, :].flatten()

            # sort ZDR
            zdr_sorted = np.sort(zdr_valid)
            ind_zdr_sorted = np.argsort(zdr_valid)

            # get the values of kdp/zh as linear interpolation of the table
            kdpzh_valid = np.interp(
                zdr_sorted, zdr_kdpzh_dict['zdr_kdpzh'][i][0, :],
                zdr_kdpzh_dict['zdr_kdpzh'][i][1, :])

            # reorder according to original order of the flat valid data array
            kdpzh_valid[ind_zdr_sorted] = kdpzh_valid

            # put it in the original matrix
            kdpzh_aux[np.logical_not(mask_aux)] = kdpzh_valid

            kdpzh[ind_ray, :] = np.reshape(
                kdpzh_aux, (len(ind_ray), np.shape(zdr)[1]))

        return refl_lin * kdpzh

    if parametrization == 'Gourley':
        zdr[zdr > 3.5] = np.ma.masked
        if zdr_kdpzh_dict['freq_band'] == 'S':
            kdpzh = 1e-5*(3.696-1.963*zdr+0.504*zdr*zdr-0.051*zdr*zdr*zdr)
        elif zdr_kdpzh_dict['freq_band'] == 'C':
            kdpzh = 1e-5*(6.746-2.970*zdr+0.711*zdr*zdr-0.079*zdr*zdr*zdr)
        elif zdr_kdpzh_dict['freq_band'] == 'X':
            kdpzh = 1e-5*(11.74-4.020*zdr-0.140*zdr*zdr+0.130*zdr*zdr*zdr)
        else:
            raise ValueError(
                'Unable to use self-consistency curves. '
                'Unknown frequency band '+zdr_kdpzh_dict['freq_band'])

        return refl_lin * kdpzh

    if parametrization == 'Wolfensberger':
        # Curve based on DSD from disdrometer data with temperature obtained
        # from disdrometer measurements. Assumed 1.0 deg elevation.
        if zdr_kdpzh_dict['freq_band'] == 'C':
            zdr[zdr > 3.5] = np.ma.masked
            kdpzh = 3.199e-5*np.ma.exp(-7.767e-1*zdr)-4.436e-6*zdr+3.464e-5
            return refl_lin * kdpzh

        raise ValueError(
            'Unable to use self-consistency curves. '
            'Unknown frequency band '+zdr_kdpzh_dict['freq_band'])

    if parametrization == 'Louf':
        # Drop shape from Brandes et al. (2002)
        if zdr_kdpzh_dict['freq_band'] == 'C':
            zdr[zdr < 0.5] = np.ma.masked
            zdr[zdr > 3.5] = np.ma.masked
            kdpzh = 1e-5*(6.607-4.577*zdr+1.577*zdr*zdr-0.23*zdr*zdr*zdr)
            return refl_lin * kdpzh

        raise ValueError(
            'Unable to use self-consistency curves. '
            'Unknown frequency band '+zdr_kdpzh_dict['freq_band'])

    if parametrization == 'Gorgucci':
        if zdr_kdpzh_dict['freq_band'] == 'C':
            zdr[zdr > 3.5] = np.ma.masked
            zdr_lin = np.ma.power(10., 0.1*zdr)
            kdpzh095 = 1.82e-4*np.power(zdr_lin, -1.28)
            return np.ma.power(refl_lin, 0.95)*kdpzh095

        raise ValueError(
            'Unable to use self-consistency curves. '
            'Unknown frequency band '+zdr_kdpzh_dict['freq_band'])

    if parametrization == 'Vaccarono':
        if zdr_kdpzh_dict['freq_band'] == 'C':
            zdr[zdr > 3.5] = np.ma.masked
            zdr_lin = np.ma.power(10., 0.1*zdr)
            kdp085zh091 = 1.77e-4*np.power(zdr_lin, -2.09)
            return np.ma.power(np.ma.power(refl_lin, 0.91)*kdp085zh091, 1./0.85)

        raise ValueError(
            'Unable to use self-consistency curves. '
            'Unknown frequency band '+zdr_kdpzh_dict['freq_band'])

    raise ValueError(
        'Unable to use self-consistency curves. '
        'Unknown parametrization '+parametrization)


def _est_sun_hit_pwr_hs(pwr, sun_hit, attg_sun, max_std, nbins_min, ind_rmin):
    """
    estimates sun hit power, standard deviation, and number and position of
    affected range bins in a ray.
    Uses Hildebrand and Sekhon 1974 to estimate the sun hit power.

    Parameters
    ----------
    pwr : 1D float array
        the power at each range bin in a ray
    sun_hit : 1D float array
        array used to flag sun hit range bins
    attg_sun : float
        attenuation suffered by the sun signal from the top of the atmosphere
        to the radar position
    max_std : float
        maximum standard deviation to consider the sun hit valid
    nbins_min : int
        minimum number of range gates with valid signal in the ray to consider
        the ray affected by a noise-like signal
    ind_rmin : int
        minimum range from which we can look for noise

    Returns
    -------
    sunpwr_dBm : float
        the estimated sun power
    sunpwr_std : float
        the standard deviation of the estimation in dB
    sunpwr_npoints : int
        the number of range gates affected by the sun hit
    nvalid : int
        the number of valid range gates
    sun_hit : 1D array
        array with flagged range bins

    """
    nvalid = len(pwr[ind_rmin:].compressed())

    if nvalid < nbins_min:
        return get_fillvalue(), get_fillvalue(), 0, nvalid, sun_hit

    pwr_toa_mw = np.ma.power(10., 0.1*(pwr[ind_rmin:]+attg_sun))
    pwr_valid = pwr_toa_mw.compressed()
    sunpwr, _, _, sunpwr_npoints = estimate_noise_hs74(pwr_valid)



    ind_sun_hits = np.argsort(pwr_valid)[0:sunpwr_npoints]
    pwr_valid = np.sort(pwr_valid)[0:sunpwr_npoints]
    sunpwr_std = np.ma.std(10.*np.ma.log10(pwr_valid))

    if sunpwr_std > max_std:
        warn('Sun hit power not valid. Standard deviation '+str(sunpwr_std) +
             ' larger than maximum expected')
        return get_fillvalue(), get_fillvalue(), 0, nvalid, sun_hit

    sunpwr_dBm = 10.*np.ma.log10(sunpwr)

    # set gates with valid sun hits to one
    is_valid = np.logical_not(np.ma.getmaskarray(sun_hit[ind_rmin:]))
    ind_valid = is_valid.nonzero()[0]
    sun_hit[ind_rmin+ind_valid[ind_sun_hits]] = 1

    return sunpwr_dBm, sunpwr_std, sunpwr_npoints, nvalid, sun_hit


def _est_sun_hit_pwr_psr(pwr, attg_sun):
    """
    estimates sun hit power

    Parameters
    ----------
    pwr : 1D float array
        the power at each range bin in a ray
    attg_sun : float
        attenuation suffered by the sun signal from the top of the atmosphere
        to the radar position

    Returns
    -------
    sunpwr_dBm : float
        the estimated sun power

    """

    pwr_toa_mw = np.ma.power(10., 0.1*(pwr+attg_sun))
    sunpwr_dBm = 10.*np.ma.log10(pwr_toa_mw)

    return sunpwr_dBm


def _est_sun_hit_pwr_ivic(pwr, sun_hit, attg_sun, pct, flat_reg_wlen,
                          flat_reg_var_max, snr_thr, npulses, ngates_min,
                          iterations):
    """
    estimates sun hit power, standard deviation, and number and position of
    affected range bins in a ray.
    Uses Ivic et al 2013 to estimate the sun hit power.

    Parameters
    ----------
    pwr : 1D float array
        the power at each range bin in a ray
    sun_hit : 1D float array
        array used to flag sun hit range bins
    attg_sun : float
        attenuation suffered by the sun signal from the top of the atmosphere
        to the radar position
    pct : float
        Point Clutter Threshold
    flat_reg_wlen : int
        Minimum number of gates that should contain a valid region. Default
        gives a size of 8 km with 83.3 m resolution
    flat_reg_var_max : float
        Maximum local variance of powers in decibels to consider the region
        as flat.
    snr_thr : float
        Threshold applied in steps 3 and 6
    npulses : int
        Number of pulses used to compute the power of the array
    ngates_min: int
        minimum number of gates with noise to consider the retrieval valid
    iterations: int
        number of iterations in step 7

    Returns
    -------
    sunpwr_dBm : float
        the estimated sun power
    sunpwr_std : float
        the standard deviation of the estimation in dB
    sunpwr_npoints : int
        the number of range gates affected by the sun hit
    nvalid : int
        the number of valid range gates
    sun_hit : 1D array
        array with flagged range bins

    """
    nvalid = len(pwr.compressed())

    pwr_toa_w = 1e-3*np.ma.power(10., 0.1*(pwr+attg_sun))
    sunpwr, _, _, inds_noise = estimate_noise_ivic13(
        pwr_toa_w, pct=pct, delay=2, flat_reg_wlen=flat_reg_wlen,
        flat_reg_var_max=flat_reg_var_max, snr_thr=snr_thr,
        npulses=npulses, ngates_min=ngates_min, iterations=iterations,
        get_noise_pos=True)

    if sunpwr is None:
        warn('No sun hit found')
        return get_fillvalue(), get_fillvalue(), 0, nvalid, sun_hit

    sunpwr_dBm = 10.*np.ma.log10(sunpwr)+30.
    pwr_toa_dBm = pwr+attg_sun
    sunpwr_std = np.ma.std(pwr_toa_dBm[inds_noise])

    # set gates with valid sun hits to one
    sun_hit[inds_noise] = 1

    return sunpwr_dBm, sunpwr_std, inds_noise.size, nvalid, sun_hit


def _est_sun_hit_zdr(zdr, sun_hit_zdr, sun_hit_h, sun_hit_v, max_std,
                     nbins_min, ind_rmin):
    """
    estimates sun hit ZDR, standard deviation, and number and position of
    affected range bins in a ray

    Parameters
    ----------
    zdr : 1D float array
        the ZDR at each range bin in a ray
    sun_hit_zdr : 1D float array
        array used to flag sun hit range bins
    sun_hit_h, sun_hit_v : 1D float array
        The position of sun hit range bins in each channel
    max_std : float
        maximum standard deviation
    nbins_min : int
        minimum number of range gates with valid signal in the ray to consider
        the ray affected by a noise-like signal
    ind_rmin : int
        minimum range from which we can look for noise

    Returns
    -------
    sunzdr : float
        the estimated sun power
    sunzdr_std : float
        the standard deviation of the estimation in dB
    sunzdr_npoints : int
        the number of range gates affected by the sun hit
    sun_hit_zdr : 1D array
        array with flagged range bins

    """
    nvalid = len(zdr[ind_rmin:].compressed())
    if nvalid < nbins_min:
        # warn('Sun hit ZDR not valid. Not enough gates with signal in ray')
        return get_fillvalue(), get_fillvalue(), 0, nvalid, sun_hit_zdr

    if sun_hit_h is None and sun_hit_v is None:
        # warn('Sun hit ZDR not valid. No sun power was detected')
        return get_fillvalue(), get_fillvalue(), 0, nvalid, sun_hit_zdr

    is_valid = np.logical_not(np.ma.getmaskarray(zdr))

    if sun_hit_h is not None:
        is_valid = np.logical_and(sun_hit_h.filled(fill_value=0), is_valid)
    if sun_hit_v is not None:
        is_valid = np.logical_and(sun_hit_v.filled(fill_value=0), is_valid)

    sunzdr_npoints = np.count_nonzero(is_valid)

    if sunzdr_npoints < 2:
        # warn('Sun hit ZDR not valid. ' +
        #      'Not enough gates with valid signal in ray')
        return get_fillvalue(), get_fillvalue(), 0, nvalid, sun_hit_zdr

    sunzdr = np.ma.mean(zdr[is_valid])
    sunzdr_std = np.ma.std(zdr[is_valid])

    if sunzdr_std > max_std:
        warn('Sun hit ZDR not valid. Standard deviation '+str(sunzdr_std) +
             ' larger than maximum expected')
        return get_fillvalue(), get_fillvalue(), 0, nvalid, sun_hit_zdr

    sun_hit_zdr[is_valid] = 1

    return sunzdr, sunzdr_std, sunzdr_npoints, nvalid, sun_hit_zdr


def _selfconsistency_kdp_phidp(
        radar, refl, zdr, phidp, zdr_kdpzh_dict, max_phidp=20.,
        smooth_wind_len=5, rhohv=None, min_rhohv=None, min_zdr=0.,
        hydro=None, filter_rain=True, doc=None, fzl=None, thickness=700.,
        parametrization='None', temp_field=None, iso0_field=None,
        temp_ref='temperature'):
    """
    Estimates KDP and PhiDP in rain from  Zh and ZDR using a selfconsistency
    relation between ZDR, Zh and KDP. Private method

    Parameters
    ----------
    radar : Radar
        radar object
    refl, zdr, phidp : ndarray 2D
        reflectivity field, differential reflectivity field and differential
        phase field. They must exist
    zdr_kdpzh_dict : dict
        dictionary containing a look up table relating ZDR with KDP/Zh for
        different elevations
    max_phidp : float
        maximum PhiDP value to consider the data valid
    smooth_wind_len : int
        length of the smoothing window for Zh and ZDR data
    rhohv : ndarray 2D
        copolar correlation field used for masking data. Optional
    min_rhohv : float
        minimum RhoHV value to consider the data valid
    min_zdr : float
        minimum ZDR value to consider the data valid
    hydro : ndarray 2D
        hydrometer classification field used for masking data. Optional
    filter_rain : Bool
        If true gates not classified as rain are going to be removed from the
        data
    doc : float
        Number of gates at the end of each ray to to remove from the
        calculation.
    fzl : float
        Freezing layer, gates above this point are not included in the
        correction.
    thickness : float
        Assumed thickness of the melting layer [m]
    parametrization : str
        The type of parametrization for the self-consistency curves. Can be
        'None', 'Gourley', 'Wolfensberger', 'Louf', 'Gorgucci' or 'Vaccarono'.
        'None' will use tables contained in zdr_kdpzh_dict.
    temp_field, iso0_field, hydro_field : str
        Field name within the radar object which represent the temperature,
        the height relative to the iso0 and the hydrometeor classification
        fields. A value of None will use the default field name as defined in
        the Py-ART configuration file. It is going to be used only if
        available.
    temp_ref : str
        the field use as reference for temperature. Can be either temperature,
        height_over_iso0 or fixed_fzl

    Returns
    -------
    kdp_sim, phidp_sim : ndarray 2D
        the KDP and PhiDP estimated fields

    """
    if (radar.instrument_parameters is not None and
            'radar_beam_width_h' in radar.instrument_parameters):
        beamwidth = (
            radar.instrument_parameters['radar_beam_width_h']['data'][0])
    else:
        warn('Unknown radar antenna beamwidth.')
        beamwidth = None

    # smooth reflectivity and ZDR
    if smooth_wind_len > 0:
        sm_refl = smooth_masked(refl, wind_len=smooth_wind_len, min_valid=1,
                                wind_type='mean')
        sm_zdr = smooth_masked(zdr, wind_len=smooth_wind_len, min_valid=1,
                               wind_type='mean')
    else:
        sm_refl = refl
        sm_zdr = zdr

    # Remove data at melting layer or above
    mask = np.ma.getmaskarray(refl)
    mask_fzl, _ = get_mask_fzl(
        radar, fzl=fzl, doc=doc, min_temp=0., max_h_iso0=0.,
        thickness=thickness, beamwidth=beamwidth, temp_field=temp_field,
        iso0_field=iso0_field, temp_ref=temp_ref)
    mask = np.logical_or(mask, mask_fzl)

    # Remove data outside of valid range of ZDR
    mask_zdr = np.logical_or(sm_zdr < min_zdr, np.ma.getmaskarray(sm_zdr))
    if parametrization == 'None':
        # Remove data with ZDR out of table values
        ele_rounded = radar.elevation['data'].astype(int)
        mask_zdr_max = np.ones((radar.nrays, radar.ngates))
        for i in range(len(zdr_kdpzh_dict['elev'])):
            ind_ray = np.where(ele_rounded == zdr_kdpzh_dict['elev'][i])[0]
            if ind_ray.size == 0:
                continue
            mask_zdr_max[ind_ray, :] = (
                sm_zdr[ind_ray, :] > zdr_kdpzh_dict['zdr_kdpzh'][i][0, -1])

        mask_zdr = np.logical_or(mask_zdr, mask_zdr_max)
    mask = np.logical_or(mask, mask_zdr)

    if max_phidp is not None:
        mask_phidp = np.logical_or(
            phidp > max_phidp, np.ma.getmaskarray(phidp))
        mask = np.logical_or(mask, mask_phidp)

    if min_rhohv is not None:
        mask_rhohv = np.logical_or(rhohv < min_rhohv,
                                   np.ma.getmaskarray(rhohv))
        mask = np.logical_or(mask, mask_rhohv)

    # Remove gates that are not rain
    if filter_rain:
        # keep only data classified as light rain (4) or rain (6)
        mask_rain = np.logical_not(np.logical_or(hydro == 4, hydro == 6))
        mask = np.logical_or(mask, mask_rain)

    corr_refl = np.ma.masked_where(mask, sm_refl)
    corr_zdr = np.ma.masked_where(mask, sm_zdr)

    kdp_sim = get_kdp_selfcons(
        corr_zdr, corr_refl, radar.elevation['data'], zdr_kdpzh_dict,
        parametrization=parametrization)
    dr = (radar.range['data'][1] - radar.range['data'][0]) / 1000.0
    phidp_sim = np.ma.cumsum(2*dr*kdp_sim, axis=1)

    return kdp_sim, phidp_sim



#! /usr/bin/env python
import pyart
import matplotlib.pyplot as plt

# plot quickly
pradar = pyart.io.read_rsl('example_sigmet_rhi.sigmet')
display = pyart.graph.RadarDisplay(pradar)
fig = plt.figure()
ax = fig.add_subplot(111)
display.plot_rhi('reflectivity_horizontal_filtered', 0, vmin=-16, vmax=64)
fig.savefig('example_sigmet_rhi.png')



"""
pyart.aux_io.rad4alp_bin_reader
===============================

Routines for putting MeteoSwiss operational radar data contained in binary
files into grid object.

.. autosummary::
    :toctree: generated/

    read_bin

"""

import os
import datetime
from warnings import warn

import numpy as np

from ..config import FileMetadata
from ..io.common import _test_arguments
from ..core.grid import Grid
from ..util import ma_broadcast_to

BIN_FIELD_NAMES = {
    'ACC': 'rainfall_accumulation',
    'ARC': 'rainfall_accumulation'
}


def read_bin(filename, additional_metadata=None, chy0=255., chx0=-160.,
             xres=1., yres=1., nx=710, ny=640, nz=1, **kwargs):

    """
    Read a MeteoSwiss operational radar data binary file.

    Parameters
    ----------
    filename : str
        Name of the file to read.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    chy0, chx0 : float
        Swiss coordinates position of the south-western point in the grid
    xres, yres : float
        resolution of each grid point [km]
    nx, ny, nz : int
        dimensions of the grid

    Returns
    -------
    grid : Grid
        Grid object containing data the data.

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    try:
        with open(filename, 'rb') as file:
            file.readline()
            data = np.fromfile(file, dtype=np.dtype('float32'), count=nx*ny*4)
            data = np.ma.masked_equal(data, -1.)
            data = np.transpose(np.reshape(data, [nx, ny], order='F'))[::-1, :]
    except EnvironmentError as ee:
        warn(str(ee))
        warn('Unable to read file '+filename)
        return None, None

    # reserved_variables = [
    #     'time', 'x', 'y', 'z',
    #     'origin_latitude', 'origin_longitude', 'origin_altitude',
    #     'point_x', 'point_y', 'point_z', 'projection',
    #     'point_latitude', 'point_longitude', 'point_altitude',
    #     'radar_latitude', 'radar_longitude', 'radar_altitude',
    #     'radar_name', 'radar_time', 'base_time', 'time_offset',
    #     'ProjectionCoordinateSystem']

    # metadata
    metadata = dict()

    filemetadata = FileMetadata('BIN', BIN_FIELD_NAMES, additional_metadata)

    # required reserved variables
    time = filemetadata('grid_time')
    origin_latitude = filemetadata('origin_latitude')
    origin_longitude = filemetadata('origin_longitude')
    origin_altitude = filemetadata('origin_altitude')
    x = filemetadata('x')
    y = filemetadata('y')
    z = filemetadata('z')

    x['data'] = 1000.*(np.arange(nx)*xres+chy0+xres/2.-600.)
    y['data'] = 1000.*(np.arange(ny)*yres+chx0+yres/2.-200.)
    z['data'] = np.array([0.])

    # Origin of LV03 Swiss coordinates
    origin_latitude['data'] = np.array([46.9524055556])
    origin_longitude['data'] = np.array([7.43958333333])
    origin_altitude['data'] = np.array([0.])

    bfile = os.path.basename(filename)
    # Time
    prod_time = (
        datetime.datetime.strptime(bfile[3:12], '%y%j%H%M') -
        datetime.timedelta(minutes=1440))
    time['units'] = 'seconds since '+prod_time.strftime('%Y-%m-%d %H:%M:%S')
    time['data'] = np.array([1440.*60.])

    # projection (Swiss Oblique Mercator)
    projection = {
        'proj': 'somerc',
        '_include_lon_0_lat_0': True
    }

    # read in the fields
    datatype = bfile[0:3]

    fields = {}
    field = filemetadata.get_field_name(datatype)
    field_dict = filemetadata(field)
    field_dict['data'] = ma_broadcast_to(data, (nz, ny, nx))
    fields[field] = field_dict

    # radar variables
    radar_latitude = None
    radar_longitude = None
    radar_altitude = None
    radar_name = None
    radar_time = None

    return Grid(
        time, fields, metadata,
        origin_latitude, origin_longitude, origin_altitude, x, y, z,
        projection=projection,
        radar_latitude=radar_latitude, radar_longitude=radar_longitude,
        radar_altitude=radar_altitude, radar_name=radar_name,
        radar_time=radar_time)



#cython: cdivision=True
#cython: boundscheck=False
#cython: nonecheck=False
#cython: wraparound=False
"""
pyart.correct._unwrap_1d
========================

.. autosummary::
    :toctree: generated/
    
    unwrap_1d

"""

from libc.math cimport M_PI


def unwrap_1d(double[::1] image, double[::1] unwrapped_image):
    '''Phase unwrapping using the naive approach.'''
    cdef:
        Py_ssize_t i
        double difference
        long periods = 0
    unwrapped_image[0] = image[0]
    for i in range(1, image.shape[0]):
        difference = image[i] - image[i - 1]
        if difference > M_PI:
            periods -= 1
        elif difference < -M_PI:
            periods += 1
        unwrapped_image[i] = image[i] + 2 * M_PI * periods



"""
==========================
Mapping (:mod:`pyart.map`)
==========================

.. current modules:: pyart.map

Py-ART has a robust function for mapping radar data from the collected radar
coordinates to Cartesian coordinates.

Mapping to grid
============

.. autosummary::
    :toctree: generated/

    grid_from_radars
    map_to_grid
    map_gates_to_grid

Examples of ROI (region of interest)
============

.. autosummary::
    :toctree: generated/

    example_roi_func_constant
    example_roi_func_dist
    example_roi_func_dist_beam

"""

from .grid_mapper import map_to_grid, grid_from_radars
from .grid_mapper import example_roi_func_constant
from .grid_mapper import example_roi_func_dist
from .grid_mapper import example_roi_func_dist_beam
from .polar_to_cartesian import polar_to_cartesian
from .polar_to_cartesian import get_earth_radius
from .gates_to_grid import map_gates_to_grid

__all__ = [s for s in dir() if not s.startswith('_')]



"""
pyart.io._sigmet_noaa_hh
========================

Functions needed for reading Sigmet files from the airborne radar located on
NOAA's Hurricane Hunter aircraft.

.. autosummary::
    :toctree: generated/

    _decode_noaa_hh_hdr
    _georeference_yprime

"""

import numpy as np
from ._sigmetfile import bin4_to_angle, bin2_to_angle


def _decode_noaa_hh_hdr(
        raw_extended_headers, filemetadata, azimuth, elevation,
        position_source='irs', heading_source='irs'):
    """
    Extract data from Sigmet extended headers produced by NOAA
    Hurricane Hunter airborne radars.

    Parameters
    ----------
    raw_extended_headers : ndarray
        Raw Sigmet extended headers.
    filemetadata : FileMetadata
        FileMetadata class from which metadata will be derived.
    azimuth : dict
        Dictionary of azimuth angles recorded in Sigmet file.
    elevation : dict
        Dictionary of elevation angles recorded in Sigmet file.
    position_source: {'irs', 'gps', 'aamps'}, optional
        Instrument from which to derive position parameters.
    heading_source: {'irs', 'aamps'}
        Instrument from which to derive heading parameters.

    Returns
    -------
    latitude : dict
        Dictionary containing latitude data and metadata.
    longitude : dict
        Dictionary containing longitude data and metadata.
    altitude : dict
        Dictionary containing altitude data and metadata.
    heading_params : dict
        Dictionary of dictionary containing aircraft heading data and
        metadata. Contains 'heading', 'roll', pitch', 'drift', 'rotation',
        'tilt' and 'georefs_applied' dictionaries.

    """
    xhdr = np.rec.frombuffer(raw_extended_headers[..., :68].tostring(),
                             dtype=list(NOAA_HH_EXTENDED_HEADER))

    # rotation and tilt from azimuth/elevation angles
    rotation = filemetadata('rotation')
    tilt = filemetadata('tilt')

    rotation_data = 90. - elevation['data'].copy()
    rotation_data[rotation_data < 0] += 360.
    rotation['data'] = rotation_data

    tilt_data = azimuth['data'].copy()
    tilt_data[tilt_data > 180] -= 360.
    tilt['data'] = tilt_data

    # airborne parameters
    heading = filemetadata('heading')
    roll = filemetadata('roll')
    pitch = filemetadata('pitch')
    drift = filemetadata('drift')

    if heading_source == 'irs':
        heading_data = bin2_to_angle(xhdr['irs_heading'])
        roll_data = bin2_to_angle(xhdr['irs_roll'])
        pitch_data = bin2_to_angle(xhdr['irs_pitch'])
        drift_data = bin2_to_angle(xhdr['irs_drift'])
    elif heading_source == 'aamps':
        heading_data = bin2_to_angle(xhdr['aamps_heading'])
        roll_data = bin2_to_angle(xhdr['aamps_roll'])
        pitch_data = bin2_to_angle(xhdr['aamps_pitch'])
        drift_data = bin2_to_angle(xhdr['aamps_drift'])
    else:
        raise ValueError('Unknown heading_source')

    heading['data'] = heading_data
    roll['data'] = roll_data
    pitch['data'] = pitch_data
    drift['data'] = drift_data

    # georeferenced azimuth and elevation
    az, elev = _georeference_yprime(
        roll_data, pitch_data, heading_data, drift_data, rotation_data,
        tilt_data)
    azimuth['data'] = az
    elevation['data'] = elev
    georefs_applied = filemetadata('georefs_applied')
    georefs_applied['data'] = np.ones(az.shape, dtype='int8')

    # positions: latitude, longitude, altitude
    latitude = filemetadata('latitude')
    longitude = filemetadata('longitude')
    altitude = filemetadata('altitude')

    if position_source == 'gps':
        lat_data = bin4_to_angle(xhdr['gps_lat'])
        lon_data = bin4_to_angle(xhdr['gps_long'])
        alt_data = xhdr['gps_alt'] / 100.
    elif position_source == 'aamps':
        lat_data = bin4_to_angle(xhdr['aamps_lat'])
        lon_data = bin4_to_angle(xhdr['aamps_long'])
        alt_data = xhdr['aamps_alt'] / 100.
    elif position_source == 'irs':
        lat_data = bin4_to_angle(xhdr['irs_lat'])
        lon_data = bin4_to_angle(xhdr['irs_long'])
        alt_data = xhdr['gps_alt'] / 100.
    else:
        raise ValueError('Invalid position_source')

    latitude['data'] = lat_data
    longitude['data'] = lon_data
    altitude['data'] = alt_data

    extended_header_params = {
        'heading': heading,
        'roll': roll,
        'pitch': pitch,
        'drift': drift,
        'rotation': rotation,
        'tilt': tilt,
        'georefs_applied': georefs_applied}
    return (latitude, longitude, altitude, extended_header_params)


def _georeference_yprime(roll, pitch, heading, drift, rotation, tilt):
    """
    Compute georeferenced azimuth and elevation angles for a Y-prime radar.

    This is the georeferencing needed for the tail doppler radar on the
    NOAA P3 aircraft.
    """
    # Adapted from Radx's SigmetRadxFile::_computeAzEl method found in
    # SigmetRadxFile.cc
    # Transforms defined in Wen-Chau Lee et al, JTech, 1994, 11, 572-578.

    # Convert to radians and use variable names from Wen-Chau Lee paper
    R = np.radians(roll)        # roll
    P = np.radians(pitch)       # pitch
    H = np.radians(heading)     # heading
    D = np.radians(drift)       # drift
    T = H + D                   # track
    theta_a = np.radians(rotation)
    tau_a = np.radians(tilt)

    # Eq. (9)
    x_t = (np.cos(theta_a + R) * np.sin(D) * np.cos(tau_a) * np.sin(P) +
           np.cos(D) * np.sin(theta_a + R) * np.cos(tau_a) -
           np.sin(D) * np.cos(P) * np.sin(tau_a))

    y_t = (-np.cos(theta_a + R) * np.cos(D) * np.cos(tau_a) * np.sin(P) +
           np.sin(D) * np.sin(theta_a + R) * np.cos(tau_a) +
           np.cos(P) * np.cos(D) * np.sin(tau_a))

    z_t = (np.cos(P) * np.cos(tau_a) * np.cos(theta_a + R) +
           np.sin(P) * np.sin(tau_a))

    # Eq. (12) and discussion after Eq. (17)
    lambda_t = np.arctan2(x_t, y_t)
    azimuth = np.fmod(lambda_t + T, 2 * np.pi)

    # Eq (17)
    elevation = np.arcsin(z_t)

    # convert to degrees and fix range
    azimuth = np.degrees(azimuth)
    azimuth[azimuth < 0] += 360.
    elevation = np.degrees(elevation)
    elevation[elevation > 180] -= 360.
    return azimuth, elevation

# NOAA Hurrican Hunter Sigmet Extended header structure
# scalar definitions
UINT16 = 'H'
INT16 = 'h'
BAM16 = 'h'
INT32 = 'i'
BAM32 = 'i'

NOAA_HH_EXTENDED_HEADER = (
    ('msecs_since_sweep_start', INT32),
    ('calib_signal_level', INT16),
    ('nbytes_in_header', INT16),
    ('__pad_1', UINT16),
    ('gps_age', UINT16),            # Time in milliseconds since last GPS Input
    ('irs_age', UINT16),            # Time in milliseconds since last IRS Input
    ('aamps_age', UINT16),          # Time in milliseconds since last
                                    # AAMPS Input
    ('gps_lat', BAM32),             # GPS latitude (BAM)
    ('gps_long', BAM32),            # GPS Longitude (BAM)
    ('gps_alt', INT32),             # GPS Altitude (cm)
    ('gps_vel_e', INT32),           # GPS Ground Speed East (cm/second)
    ('gps_vel_n', INT32),           # GPS Ground Speed North (cm/second)
    ('gps_vel_v', INT32),           # GPS Ground Speed Up (cm/second)
    ('irs_lat', BAM32),             # IRS latitude (BAM)
    ('irs_long', BAM32),            # IRS Longitude (BAM)
    ('irs_vel_e', INT32),           # IRS Ground Speed East (cm/second)
    ('irs_vel_n', INT32),           # IRS Ground Speed North (cm/second)
    ('irs_vel_v', INT32),           # IRS Ground Speed Up (cm/second)
    ('irs_pitch', BAM16),           # IRS Pitch (BAM)
    ('irs_roll', BAM16),            # IRS Roll (BAM)
    ('irs_heading', BAM16),         # IRS Heading (BAM)
    ('irs_drift', BAM16),           # IRS Drift (BAM)
    ('irs_tru_track', BAM16),       # IRS True Track (BAM)
    ('irs_pitch_r', BAM16),         # IRS Pitch rate (BAM/sec)
    ('irs_roll_r', BAM16),          # IRS Roll rate (BAM/sec)
    ('irs_yaw_r', BAM16),           # IRS Yaw rate (BAM/sec)
    ('irs_wind_vel', INT32),        # IRS Wind speed (cm/second)
    ('irs_wind_dir', BAM16),        # IRS Wind direction (BAM)
    ('__pad_2', UINT16),
    ('aamps_lat', BAM32),           # AAMPS latitude (BAM)
    ('aamps_long', BAM32),          # AAMPS Longitude (BAM)
    ('aamps_alt', INT32),           # AAMPS Altitude (cm)
    ('aamps_ground_vel', INT32),    # AAMPS Ground Speed East (cm/second)
    ('aamps_time_stamp', INT32),    # AAMPS Timestamp in UTC
                                    # (seconds since the epoch)
    ('aamps_vel_v', INT32),         # AAMPS Vertical Velocity (cm/second)
    ('aamps_pitch', BAM16),         # AAMPS Pitch (BAM)
    ('aamps_roll', BAM16),          # AAMPS Roll (BAM)
    ('aamps_heading', BAM16),       # AAMPS Heading (BAM)
    ('aamps_drift', BAM16),         # AAMPS Drift (BAM)
    ('aamps_track', BAM16),         # AAMPS Track (BAM)
    ('__pad_4', UINT16),
    ('aamps_radalt_val', INT32),    # AAMPS Radar Altitude (cm)
    ('aamps_wind_vel', INT32),      # AAMPS Wind Speed (cm/second)
    ('aamps_wind_dir', BAM16),      # AAMPS Wind direction (BAM)
    ('__pad_5', UINT16),
    ('aamps_wind_vel_v', INT32),    # AAMPS Wind Speed Up (cm/second)
)



"""
pyart.retrieve.gecsx_functions
========================

Cython routines for the gecsx visibility processing

.. autosummary::
    :toctree: generated/

    grid_displacement_pc
    grid_shift

"""
# "cimport" is used to import special compile-time information
# about the numpy module (this is stored in a file numpy.pxd which is
# currently part of the Cython distribution).
cimport numpy as np
import numpy as np

# We now need to fix a datatype for our arrays. I've used the variable
# DTYPE for this, which is assigned to the usual NumPy runtime
# type info object.
DTYPE = np.double

# "ctypedef" assigns a corresponding compile-time type to DTYPE_t. For
# every type in the numpy module there's a corresponding compile-time
# type with a _t-suffix.
ctypedef np.double_t DTYPE_t

def argmin_abs(np.ndarray[DTYPE_t, ndim=1] vec, double elem):
    cdef double min = np.inf
    cdef int minidx = 0
    cdef int i
    for i in range(len(vec)):
        if abs(vec[i] - elem) < min:
            min = vec[i]
            minidx = i
    return minidx

def vis_weighting(np.ndarray[DTYPE_t, ndim=1] daz_vec, 
                  np.ndarray[DTYPE_t, ndim=1] del_vec,
                  np.ndarray[DTYPE_t, ndim=1] daz_area,
                  np.ndarray[DTYPE_t, ndim=2] ant_weight,
                  np.ndarray[DTYPE_t, ndim=1] minvisvals,
                  double el):
    
    cdef int ndaz = len(daz_vec)
    cdef int ndel = len(del_vec)
    cdef int idxmin = 0
    cdef double vis = 0
    cdef int idaz, idel
    for idaz in range(ndaz):
        mina = argmin_abs(daz_area, daz_vec[idaz])
        for idel in range(ndel):
            if minvisvals[mina] <= el + del_vec[idel]:
                vis += ant_weight[idaz, idel]
    return vis



""" Shared extensions in _rsl_interface, used by _fourdd_interface. """


cimport _rsl_h

cdef class _RslVolume:
    cdef _rsl_h.Volume * _Volume
    cdef int _dealloc
    cdef load(self, _rsl_h.Volume * Volume)
    cdef _prtmode(self, _rsl_h.Ray_header h)



""" Unit Tests for Py-ART's output_to_geotiff.py module. """

import warnings

import numpy as np
import pytest

import pyart


# TODO : inspect the output file to verify their contents, currently only the
# fact that something was written is confirmed

def test__get_rgb_values_nan():
    data = np.arange(10, dtype='float32')
    data[5] = np.nan
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        rarr, barr, garr = pyart.io.output_to_geotiff._get_rgb_values(
            data, 0, 10, None, 'jet')
    assert np.isnan(rarr[5])
    assert np.isnan(barr[5])
    assert np.isnan(garr[5])


def test_raise_missingoptionaldepedency():
    backup = bool(pyart.io.output_to_geotiff.IMPORT_FLAG)
    pyart.io.output_to_geotiff.IMPORT_FLAG = False
    grid = make_tiny_grid()
    pytest.raises(
        pyart.exceptions.MissingOptionalDependency,
        pyart.io.write_grid_geotiff, grid, 'test.foo', 'reflectivity')
    pyart.io.output_to_geotiff.IMPORT_FLAG = backup


def make_tiny_grid():
    """ Make a tiny grid. """
    grid_shape = (2, 10, 8)
    grid_limits = ((0, 500), (-400000, 400000), (-300000, 300000))
    grid = pyart.testing.make_empty_grid(grid_shape, grid_limits)
    fdata = np.zeros((2, 10, 8), dtype='float32')
    fdata[:, 2:-2, 1:-1] = 10.
    fdata[:, 3:-3, 2:-2] = 20.
    fdata[:, 4:-4, 3:-3] = 30.
    fdata[1] += 5
    rdic = {
        'data': fdata,
        'long_name': 'reflectivity',
        'units': 'dBz'}
    grid.fields = {'reflectivity': rdic}
    return grid


@pytest.mark.skipif(not pyart.io.output_to_geotiff.IMPORT_FLAG,
                    reason="GDAL is not installed.")
def test_write_grid_geotiff_tif_single_channel():
    grid = make_tiny_grid()
    with pyart.testing.InTemporaryDirectory():
        pyart.io.write_grid_geotiff(grid, 'test.foo', 'reflectivity')
        # check that something was written to the file
        with open('test.foo', 'rb') as f:
            assert len(f.read()) > 0


@pytest.mark.skipif(not pyart.io.output_to_geotiff.IMPORT_FLAG,
                    reason="GDAL is not installed.")
def test_write_grid_geotiff_tif_rgb_with_missing():
    grid = make_tiny_grid()
    with pyart.testing.InTemporaryDirectory():
        pyart.io.write_grid_geotiff(
            grid, 'test.foo', 'reflectivity', rgb=True)
        # check that something was written to the file
        with open('test.foo', 'rb') as f:
            assert len(f.read()) > 0


@pytest.mark.skipif(not pyart.io.output_to_geotiff.IMPORT_FLAG,
                    reason="GDAL is not installed.")
def test_write_grid_geotiff_tif_level():
    grid = make_tiny_grid()
    with pyart.testing.InTemporaryDirectory():
        pyart.io.write_grid_geotiff(grid, 'test.foo', 'reflectivity', level=1)
        # check that something was written to the file
        with open('test.foo', 'rb') as f:
            assert len(f.read()) > 0


@pytest.mark.skipif(not pyart.io.output_to_geotiff.IMPORT_FLAG,
                    reason="GDAL is not installed.")
def test_write_grid_geotiff_tif_warp():
    grid = make_tiny_grid()
    with pyart.testing.InTemporaryDirectory():
        pyart.io.write_grid_geotiff(
            grid, 'test.foo', 'reflectivity', warp=True)
        # check that something was written to the file
        with open('test.foo', 'rb') as f:
            assert len(f.read()) > 0


@pytest.mark.skipif(not pyart.io.output_to_geotiff.IMPORT_FLAG,
                    reason="GDAL is not installed.")
def test_write_grid_geotiff_tif_missing_suffix():
    grid = make_tiny_grid()
    with pyart.testing.InTemporaryDirectory():
        pyart.io.write_grid_geotiff(grid, 'test', 'reflectivity')
        # suffix missing, should be added to filename
        # check that something was written to the file
        with open('test.tif', 'rb') as f:
            assert len(f.read()) > 0


@pytest.mark.skipif(not pyart.io.output_to_geotiff.IMPORT_FLAG,
                    reason="GDAL is not installed.")
def test_write_grid_geotiff_sld():
    grid = make_tiny_grid()
    with pyart.testing.InTemporaryDirectory():
        pyart.io.write_grid_geotiff(
            grid, 'test.foo', 'reflectivity', sld=True)
        # check that something was written to the file
        with open('test.foo', 'rb') as f:
            assert len(f.read()) > 0


@pytest.mark.skipif(not pyart.io.output_to_geotiff.IMPORT_FLAG,
                    reason="GDAL is not installed.")
def test_write_grid_geotiff_missing_field():
    grid = make_tiny_grid()
    pytest.raises(
        KeyError, pyart.io.write_grid_geotiff, grid, 'test.foo', 'foobar')
